use core::{
    any::Any,
    convert::TryInto,
    fmt::{Error, Write},
};

use crate::{drivers::UartDevice, irq::wait_for_irq_and_run_next, sync::UPSafeCell};
use alloc::collections::VecDeque;

pub struct Ns1665a {
    base_address: usize,
    buffer: UPSafeCell<VecDeque<u8>>,
}

impl Ns1665a {
    pub fn new() -> Self {
        let mut ns = Self {
            base_address: 0x10000000,
            buffer: unsafe { UPSafeCell::new(VecDeque::new()) },
        };
        ns.init();
        ns
    }

    pub fn init(&mut self) {
        let ptr = self.base_address as *mut u8;
        unsafe {
            let lcr: u8 = (1 << 0) | (1 << 1);
            ptr.add(3).write_volatile(lcr);
            ptr.add(2).write_volatile(1 << 0);
            ptr.add(1).write_volatile(1 << 0);
            let divisor: u16 = 592;
            let divisor_least: u8 = (divisor & 0xff).try_into().unwrap();
            let divisor_most: u8 = (divisor >> 8).try_into().unwrap();
            ptr.add(3).write_volatile(lcr | 1 << 7);
            ptr.add(0).write_volatile(divisor_least);
            ptr.add(1).write_volatile(divisor_most);
            ptr.add(3).write_volatile(lcr);
        }
    }
}

impl UartDevice for Ns1665a {
    fn getchar(&self) -> Option<u8> {
        let mut buffer = self.buffer.exclusive_access();
        let len = buffer.len();
        drop(buffer);
        if len == 0 {
            wait_for_irq_and_run_next(10);
        }
        let mut buffer = self.buffer.exclusive_access();
        assert!(buffer.len() > 0);
        buffer.pop_front()
    }

    fn putchar(&self, ch: u8) {
        let ptr = self.base_address as *mut u8;
        unsafe {
            ptr.add(0).write_volatile(ch);
        }
    }

    fn handler_interrupt(&self) {
        let ptr = self.base_address as *mut u8;
        let mut buffer = self.buffer.exclusive_access();
        unsafe {
            if ptr.add(5).read_volatile() & 1 != 0 {
                buffer.push_back(ptr.add(0).read_volatile());
            }
        }
    }
}
