use crate::{
    fs::{Dirent, File, Kstat, VDirent},
    sync::UPSafeCell,
    timer::{self, TimeSpec, TimeVal},
};
use alloc::{string::String, vec::Vec};
use core::{cmp, ptr::NonNull};
use k210_pac::{aes::en, wdt0::cr};
use log::info;
const MAX_FILE_SIZE: u32 = core::u32::MAX;
use super::{
    alloc_cluster, cluster_to_offset,
    dir_entry::{DirEntry, DirEntryEditor, DirFileEntry},
    io::{Error, IoBase, Read, Seek, SeekFrom, Write},
    sdcard::BlkCacheManager,
    time::DateTime,
    FATFS,
};

pub struct FileEntry {
    pub name: String,
    pub first_cluster: Option<u32>,
    pub current_cluster: Option<u32>,
    pub pos: u64,
    pub entry: DirEntryEditor,
    pub disk: UPSafeCell<BlkCacheManager>,
    pub range: (u64, u64),
    pub atime: TimeSpec,
    pub mtime: TimeSpec,
}

impl FileEntry {
    pub fn from(entry: DirFileEntry, entry_pos: u64, name: String) -> Self {
        let pos = 0;
        let abs_start_pos = match entry.first_cluster() {
            Some(x) => cluster_to_offset(x),
            None => 0,
        };
        let range = (abs_start_pos, abs_start_pos + entry.size());

        let mut disk = BlkCacheManager::from(abs_start_pos as usize, entry.size() as usize);
        disk.seek(SeekFrom::Start(abs_start_pos)).unwrap();
        let disk = unsafe { UPSafeCell::new(disk) };
        let entry = DirEntryEditor::new(entry, entry_pos);
        let first_cluster = entry.data.first_cluster();
        let current_cluster = entry.data.first_cluster();
        Self {
            pos,
            entry,
            disk,
            range,
            first_cluster,
            current_cluster,
            name,
            atime: unsafe { core::mem::zeroed() },
            mtime: unsafe { core::mem::zeroed() },
        }
    }
    pub fn size(&self) -> u64 {
        self.entry.data.size()
    }
    pub fn set_first_cluster(&mut self, cluster: u32) {
        self.first_cluster = Some(cluster);
        self.entry.inner().set_first_cluster(self.first_cluster);
    }
    pub fn update_dir_entry_after_write(&mut self) {
        self.entry.set_size(self.pos as u32);
    }
    pub fn stat(&self, stat: &mut Kstat) {
        stat.st_dev = 1;
        stat.sd_ino = 1;
        stat.st_mode = 100000;
        stat.st_nlink = 1;
        stat.st_uid = 0;
        stat.st_gid = 0;
        stat.st_rdev = 1;
        stat.st_size = self.size() as i64;
        stat.st_blksize = 512;
        stat.st_blocks = 0;
        let access = &self.atime;
        let create = self.entry.data.created();
        let modify = &self.mtime;
        stat.st_atime_sec = access.sec as i64;
        stat.st_atime_nsec = access.nsec as i64;
        stat.st_mtime_sec = modify.sec as i64;
        stat.st_mtime_nsec = modify.nsec as i64;
        stat.st_ctime_sec = create.time.sec() as i64;
        stat.st_ctime_nsec = create.time.msec() as i64 * 1000;
    }

    pub fn utimens(&mut self, atime: &TimeSpec, mtime: &TimeSpec) {
        if atime.nsec != 0x3ffffffe {
            let time = if atime.nsec == 0x3fffffff {
                timer::set_timespec(&mut self.atime);
                DateTime::now()
            } else {
                self.atime = *atime;
                DateTime::from_TimeSpec(atime)
            };
            self.entry.data.set_accessed(time.date);
        }
        if mtime.nsec != 0x3ffffffe {
            let time = if mtime.nsec == 0x3fffffff {
                timer::set_timespec(&mut self.mtime);
                DateTime::now()
            } else {
                self.mtime = *mtime;
                DateTime::from_TimeSpec(mtime)
            };
            self.entry.data.set_modified(time)
        }
    }
}

impl IoBase for FileEntry {
    type Error = ();
}

impl Seek for FileEntry {
    fn seek(&mut self, pos: SeekFrom) -> Result<u64, Self::Error> {
        let disk_offset = match pos {
            SeekFrom::Start(x) => {
                let offset = self.range.0 + x;
                self.pos = x as u64;
                offset
            }
            SeekFrom::End(x) => {
                let offset = self.range.1 as i64 + x;
                self.pos = (self.size() as i64 + x) as u64;
                offset as u64
            }
            SeekFrom::Current(x) => {
                let offset = x + self.pos as i64;
                self.range.0 + offset as u64
            }
        };
        self.disk
            .inner
            .borrow_mut()
            .seek(SeekFrom::Start(disk_offset))
            .unwrap();

        Ok(self.pos)
    }
}

impl Read for FileEntry {
    fn read(&mut self, buf: &mut [u8]) -> Result<usize, Self::Error> {
        let mut disk = self.disk.inner.borrow_mut();
        let off = self.size() - self.pos;
        disk.seek(SeekFrom::Start(
            cluster_to_offset(self.entry.inner().first_cluster().unwrap()) + self.pos,
        ))
        .unwrap();
        match disk.read(buf) {
            Ok(len) => {
                let len = len as u64;
                if off <= len {
                    self.pos += off;
                    Ok(off as usize)
                } else {
                    self.pos += len;
                    Ok(len as usize)
                }
            }
            Err(x) => Err(x),
        }
    }
}

impl Write for FileEntry {
    fn write(&mut self, buf: &[u8]) -> Result<usize, Self::Error> {
        let cluster_size = FATFS.cluster_size() as u64;
        let offset_in_cluster = self.pos % cluster_size;
        let bytes_left_in_cluster = (cluster_size - offset_in_cluster) as usize;
        let bytes_left_until_max_file_size = (MAX_FILE_SIZE - self.pos as u32) as usize;
        let write_size = cmp::min(buf.len(), bytes_left_in_cluster);
        let write_size = cmp::min(write_size, bytes_left_until_max_file_size);
        if write_size == 0 {
            return Ok(0);
        }
        let current_cluster = if self.pos % cluster_size == 0 {
            let next_cluster = match self.current_cluster {
                Some(n) => return Err(()),
                None => self.first_cluster,
            };
            if let Some(n) = next_cluster {
                n
            } else {
                let new_cluster = alloc_cluster(self.current_cluster, false).unwrap();
                if self.first_cluster.is_none() {
                    self.set_first_cluster(new_cluster);
                }
                new_cluster
            }
        } else {
            match self.current_cluster {
                Some(n) => n,
                None => panic!("Offset inside cluster but no cluster allocated"),
            }
        };
        let offset_in_fs = cluster_to_offset(current_cluster) + u64::from(offset_in_cluster);
        let written_bytes = {
            let mut disk = BlkCacheManager::new();
            disk.seek(SeekFrom::Start(offset_in_fs))?;
            disk.write(&buf[..write_size])?
        };
        if written_bytes == 0 {
            return Ok(0);
        }
        self.pos += written_bytes as u64;
        self.current_cluster = Some(current_cluster);
        self.update_dir_entry_after_write();
        self.entry.flush();
        Ok(written_bytes)
    }

    fn flush(&mut self) -> Result<(), Self::Error> {
        self.disk.exclusive_access().flush()
    }
}

pub struct ArrFile {
    pub offset: usize,
    pub v:Vec<u8>
}

pub enum Inode {
    File(FileEntry),
    Dir(DirEntry),
    VDir(VDirent),
    Arr(ArrFile)
}

impl Inode {
    pub fn ls(&mut self) {
        match self {
            Inode::Dir(dir) => dir.ls(),
            _ => {}
        }
    }

    pub fn is_dir(&self) -> bool {
        match self {
            Inode::Dir(dir) => dir.is_dir(),
            _ => false
            
        }
    }
    pub fn is_file(&self) -> bool {
        match self {
            Inode::VDir(_) => false,
            Inode::File(_) => true,
            Inode::Dir(dir) => !dir.is_dir(),
            Inode::Arr(_) => true,
        }
    }

    pub fn create(&mut self, name: &str, isdir: bool) -> Option<Inode> {
        match self {
            Inode::Dir(dir) => match isdir {
                true => {
                    if let Ok(file) = dir.create_dir(name) {
                        return Some(file);
                    }
                }
                false => {
                    if let Ok(file) = dir.create_file(name) {
                        return Some(file);
                    }
                }
            },
            _ => {}
        }
        None
    }

    pub fn open(&mut self, name: &str, isdir: bool) -> Option<Inode> {
        match self {
            Inode::Dir(dir) => match isdir {
                true => {
                    if let Ok(dir) = dir.open_dir(name) {
                        return Some(dir);
                    }
                }
                false => {
                    if let Ok(file) = dir.open_file(name) {
                        return Some(file);
                    }
                }
            },
            _ => {}
        }
        None
    }

    pub fn read(&mut self, offset: usize, buf: &mut [u8]) -> usize {
        match self {
            Inode::VDir(_) => 0,
            Inode::File(file) => {
                file.seek(SeekFrom::Start(offset as u64)).unwrap();
                file.read(buf).unwrap()
            }
            Inode::Dir(_) => 0,
            Inode::Arr(arr) => {
                arr.offset = offset;
                let mut len = 0;
                for i in 0..buf.len() {
                    buf[i] = arr.v[arr.offset + i];
                    len += 1;
                }
                len
            }
        }
    }

    pub fn write(&mut self, offset: usize, buf: &mut [u8]) -> usize {
        match self {
            Inode::VDir(_) => 0,
            Inode::File(file) => {
                file.seek(SeekFrom::Start(offset as u64)).unwrap();
                file.write(buf).unwrap()
            }
            Inode::Dir(_) => 0,
            Inode::Arr(arr) => {
                arr.offset = offset;
                let mut len = 0;
                for i in 0..buf.len() {
                    arr.v[arr.offset + i] = buf[i];
                    len += 1;
                }
                len
            }
        }
    }

    pub fn read_all(&mut self, offset: usize) -> Vec<u8> {
        match self {
            Inode::VDir(dir) => panic!("Not File {} ", dir.name()),
            Inode::File(file) => {
                let mut v = Vec::new();
                file.seek(SeekFrom::Start(offset as u64)).unwrap();
                file.read_to_end(&mut v).unwrap();
                v
            }
            Inode::Dir(dir) => panic!("Not File {} ", dir.file_name()),
            Inode::Arr(arr) => arr.v.clone(),
            
        }
    }

    pub fn file_name(&self) -> String {
        match self {
            Inode::VDir(vdir) => vdir.name(),
            Inode::File(file) => String::clone(&file.name),
            Inode::Dir(dir) => dir.file_name(),
            _ => String::from("_")
        }
    }
    pub fn seek(&mut self, offset: SeekFrom) -> usize {
        match self {
            Inode::VDir(_) => 0,
            Inode::File(file) => file.seek(offset).unwrap() as usize,
            Inode::Dir(dir) => 0,
            Inode::Arr(arr) => {
                match offset {
                    SeekFrom::Start(offset) => {
                        arr.offset = offset as usize;
                    }
                    SeekFrom::End(offset) => {
                        arr.offset = arr.v.len() as usize - offset as usize;
                    }
                    SeekFrom::Current(offset) => {
                        arr.offset += offset as usize;
                    }
                }
                arr.offset
            }
        }
    }

    pub fn remove(&mut self, path: &str) -> bool {
        match self {
            Inode::VDir(_) => false,
            Inode::File(file) => false,
            Inode::Dir(dir) => dir.remove(path).is_ok(),
            Inode::Arr(_) => false
        }
    }

    pub fn stat(&mut self, stat: &mut Kstat) {
        match self {
            Inode::VDir(_) => {}
            Inode::File(file) => file.stat(stat),
            Inode::Dir(dir) => dir.stat(stat),
            Inode::Arr(_) => {}
        }
    }
    pub fn getdents(&mut self, dirent: &mut Dirent) -> isize {
        match self {
            Inode::VDir(_) => -1,
            Inode::File(_) => -1,
            Inode::Dir(dir) => {
                dir.seek(SeekFrom::Start(0)).unwrap();
                let current = dir.dirents;
                let mut nread = 0;
                dirent.d_ino = current as usize;
                dirent.d_off = 0;
                dirent.d_reclen = (dir.dirents - current) as u16;
                unsafe {
                    let d_name = core::slice::from_raw_parts_mut(
                        dirent.d_name.as_ptr() as *mut u8,
                        dir.file_name().len(),
                    );
                    d_name.clone_from_slice(dir.file_name().as_bytes());
                }
                for e in dir {
                    nread += 1;
                }
                nread
            },
            Inode::Arr(_) => -1
        }
    }

    pub fn utimens(&mut self, atime: &TimeSpec, ctime: &TimeSpec) {
        match self {
            Inode::File(file) => file.utimens(atime, ctime),
            _ => {}
        }
    }
}

impl Drop for Inode {
    fn drop(&mut self) {
        match self {
            Inode::File(file) => file.flush().unwrap(),
            _ => {}
        }
    }
}
