use crate::drivers::{BlockDevice, BLOCK_DEVICE};
use crate::fatfs::file::Inode;
use crate::fatfs::io::SeekFrom;
use crate::fatfs::root_dir;
use crate::fs::File;
use crate::mm::UserBuffer;
use crate::sync::{UPSafeCell, SpinMutex};
use alloc::string::String;
use alloc::sync::Arc;
use alloc::vec::Vec;

pub struct OSInode {
    readable: bool,
    writable: bool,
    inner: SpinMutex<OSInodeInner>,
}
pub struct OSInodeInner {
    offset: usize,
    inode: Arc<UPSafeCell<Inode>>,
}

impl OSInode {
    pub fn new(readable: bool, writable: bool, inode: Arc<UPSafeCell<Inode>>) -> Self {
        Self {
            readable,
            writable,
            inner: SpinMutex::new(OSInodeInner { offset: 0, inode }),
        }
    }

    pub fn read_all(&self) -> Vec<u8> {
        let inner = self.inner.lock().unwrap();
        let offset = inner.offset;
        let v = inner.inode.inner.borrow_mut().read_all(offset);
        v
    }

    pub fn seek(&self, seek: usize) {
        let mut inner = self.inner.lock().unwrap();
        let mut inode = inner.inode.exclusive_access();
        let offset = inode.seek(SeekFrom::Start(seek as u64));
        drop(inode);
        inner.offset = offset;
    }

    pub fn read(&self, buf: &mut [u8]) -> usize {
        let mut inner = self.inner.lock().unwrap();
        let offset = inner.offset;
        let mut inode = inner.inode.exclusive_access();
        let len = inode.read(offset, buf);
        drop(inode);
        inner.offset = offset + len;
        len
    }
    pub fn close(&self) {}
}

bitflags! {
    pub struct OpenFlags: u32 {
        const RDONLY = 0x000;
        const WRONLY = 0x001;
        const RDWR = 0x002;
        const CREATE = 0x40;
        const TRUNC = 1 << 10;
        const DIRECTORY = 0x0200000;
        const DIR = 0x040000;
        const FILE = 0x100000;
        const NONBLOK = 04000;
        const LARGEFILE = 0100000;
        const CLOEXEC = 02000000;
    }
}

impl OpenFlags {
    pub fn read_write(&self) -> (bool, bool) {
        if self.is_empty() {
            (true, false)
        } else if self.contains(Self::WRONLY) {
            (false, true)
        } else {
            (true, true)
        }
    }
}

pub fn root() -> Arc<OSInode> {
    Arc::new(OSInode::new(
        true,
        true,
        Arc::new(unsafe { UPSafeCell::new(root_dir()) }),
    ))
}

pub fn open_file(path: &str, flags: OpenFlags) -> Option<Arc<OSInode>> {
    let (readable, writable) = flags.read_write();
    if flags.contains(OpenFlags::CREATE) {
        let file = root_dir().create(path, false).unwrap();
        Some(Arc::new(OSInode::new(
            readable,
            writable,
            Arc::new(unsafe { UPSafeCell::new(file) }),
        )))
    } else {
        if let Some(file) = root_dir().open(path, false) {
            Some(Arc::new(OSInode::new(
                readable,
                writable,
                Arc::new(unsafe { UPSafeCell::new(file) }),
            )))
        } else {
            println!("None {}", path);
            None
        }
    }
}

impl File for OSInode {
    fn readable(&self) -> bool {
        self.readable
    }
    fn writable(&self) -> bool {
        self.writable
    }

    fn open(&self, name: &str, read: bool, write: bool, isdir: bool) -> Option<Arc<OSInode>> {
        let inner = self.inner.lock().unwrap();
        let mut inner = inner.inode.exclusive_access();
        if let Some(inode) = inner.open(name, isdir) {
            let os_inode = OSInode::new(read, write, Arc::new(unsafe { UPSafeCell::new(inode) }));
            Some(Arc::new(os_inode))
        } else {
            None
        }
    }

    fn seek(&self, offset: SeekFrom) -> usize {
        let mut inner = self.inner.lock().unwrap();
        let offset = inner.inode.exclusive_access().seek(offset);
        inner.offset = offset;
        offset
    }

    fn read(&self, mut buf: UserBuffer) -> usize {
        let mut inner = self.inner.lock().unwrap();
        let mut total_read_size = 0usize;
        for slice in buf.buffers.iter_mut() {
            let offset = inner.offset;
            let len = inner.inode.exclusive_access().read(offset, *slice);
            if len == 0 {
                break;
            }
            inner.offset += len;
            total_read_size += len;
        }
        total_read_size
    }
    fn write(&self, mut buf: UserBuffer) -> usize {
        let mut inner = self.inner.lock().unwrap();

        let mut total_write_size = 0usize;
        for slice in buf.buffers.iter_mut() {
            let offset = inner.offset;
            let len = inner.inode.inner.borrow_mut().write(offset, *slice);
            if len == 0 {
                break;
            }
            inner.offset += len;
            total_write_size += len;
        }
        total_write_size
    }
    fn create(&self, name: &str, read: bool, write: bool, isdir: bool) -> Option<Arc<OSInode>> {
        let inner = self.inner.lock().unwrap();
        let mut inner = inner.inode.exclusive_access();
        if let Some(inode) = inner.create(name, isdir) {
            let os_inode = OSInode::new(read, write, Arc::new(unsafe { UPSafeCell::new(inode) }));
            Some(Arc::new(os_inode))
        } else {
            None
        }
    }

    fn kstat(&self, stat: &mut Kstat) {
        self.inner
            .lock().unwrap()
            .inode
            .inner
            .borrow_mut()
            .stat(stat)
    }

    fn remove(&self, path: &str) -> bool {
        self.inner
        .lock()
        .unwrap()
            .inode
            .inner
            .borrow_mut()
            .remove(path)
    }

    fn name(&self) -> String {
        self.inner
            .lock()
            .unwrap()
            .inode
            .inner
            .borrow_mut()
            .file_name()
    }
    fn getdents(&self, dirent: &mut Dirent) -> isize {
        self.inner
            .lock()
            .unwrap()
            .inode
            .exclusive_access()
            .getdents(dirent)
    }
    fn utimens(&self, ctime: &crate::timer::TimeSpec, mtime: &crate::timer::TimeSpec) {
        self.inner
            .lock()
            .unwrap()
            .inode
            .exclusive_access()
            .utimens(ctime, mtime)
    }
}

#[repr(C)]
#[derive(Debug, Default)]
pub struct Kstat {
    pub st_dev: u64,
    pub sd_ino: u64,
    pub st_mode: u32,
    pub st_nlink: u32,
    pub st_uid: u32,
    pub st_gid: u32,
    pub st_rdev: u64,
    pub __pad: u64,
    pub st_size: i64,
    pub st_blksize: u32,
    pub __pad2: i32,
    pub st_blocks: u64,
    pub st_atime_sec: i64,
    pub st_atime_nsec: i64,
    pub st_mtime_sec: i64,
    pub st_mtime_nsec: i64,
    pub st_ctime_sec: i64,
    pub st_ctime_nsec: i64,
    pub __unused: [i32; 2],
}

impl Kstat {
    pub fn as_bytes(&self) -> &[u8] {
        let size = core::mem::size_of::<Self>();
        unsafe { core::slice::from_raw_parts(self as *const _ as usize as *const u8, size) }
    }
}

#[repr(C)]
pub struct Dirent {
    pub d_ino: usize,
    pub d_off: isize,
    pub d_reclen: u16,
    pub d_type: u8,
    pub d_name: [u8; 0],
}
