use alloc::{string::String, sync::Arc, vec::Vec, collections::VecDeque};

use crate::{fatfs::io::SeekFrom, sync::SpinMutex};

use super::{File, Kstat, OSInode};

pub fn vfs_read(inode: Arc<OSInode>, buf: &mut Vec<u8>, offset: usize, len: usize) {
    assert!(buf.len() >= len);
    inode.seek(offset);
    inode.read(buf);
}

pub struct VDirent {
    pub name: String,
    pub dirs: Vec<Arc<dyn File + Send + Sync>>,
}

impl File for VDirent {
    fn readable(&self) -> bool {
        true
    }

    fn writable(&self) -> bool {
        true
    }

    fn name(&self) -> alloc::string::String {
        self.name.clone()
    }
    fn kstat(&self, stat: &mut Kstat) {
        stat.st_mode = 0x21b6;
    }
}

pub struct NullDev {}

impl File for NullDev {
    fn readable(&self) -> bool {
        true
    }

    fn writable(&self) -> bool {
        true
    }

    fn name(&self) -> alloc::string::String {
        String::from("/dev/null")
    }
    fn kstat(&self, stat: &mut Kstat) {
        stat.st_mode = 0o20000;
    }
}

pub struct ZeroDev {

}

impl File for ZeroDev {
    fn readable(&self) -> bool {
        true
    }

    fn writable(&self) -> bool {
        true
    }

    fn name(&self) -> alloc::string::String {
        String::from("/dev/zero")
    }
    fn kstat(&self, stat: &mut Kstat) {
        stat.st_mode = 0o20000;
    }
    fn read(&self, buf: crate::mm::UserBuffer) -> usize {
        let mut len = 0;
        for v in buf.buffers {
            for ch in v {
                *ch = 0;
                len += 1;
            }
        }
        len
    }
}
#[derive(Debug)]
pub struct SockAddr{
    sa_family_t: u16,
    sa_data: [u8; 14],
}

pub struct Socket {
    data: SpinMutex<VecDeque<u8>>
}

impl Socket {
    pub fn new() -> Socket {
        Socket {
            data: SpinMutex::new(VecDeque::new())
        }
    }
}

impl File for Socket {
    fn readable(&self) -> bool {
        true
    }

    fn writable(&self) -> bool {
        true
    }

    fn name(&self) -> alloc::string::String {
        String::from("socket")
    }
    fn kstat(&self, stat: &mut Kstat) {
        stat.st_mode = 0o20000;
    }
    fn bind(&self, addr: SockAddr, addrlen: isize) -> isize {
        0
    }
    fn read(&self, buf: crate::mm::UserBuffer) -> usize {
        let mut len = 0;
        let mut data = self.data.lock().unwrap();
        println!("Data Len {}",data.len());
        for v in buf.buffers {
            for ch in v {
                if data.len() > 0 {
                    *ch = data.pop_front().unwrap();
                    len += 1;
                } else {
                    break;
                }
            }
        }
        len
    }
    fn write(&self, buf: crate::mm::UserBuffer) -> usize {
        let mut len = 0;
        let mut data = self.data.lock().unwrap();
        println!("Write {}",buf.len());
        for v in buf.buffers {
            for ch in v {
                data.push_back(*ch);
                len += 1;
            }
        }
        len
    }
}