use core::fmt::Debug;

use crate::{
    drivers::{PlicDevice, BLOCK_DEVICE, PLIC_DRIVE, UART_DEVICE},
    sync::SpinMutex,
    task::{
        add_task, current_hartid, schedule, take_current_task, TaskContext, TaskControlBlock,
        TaskStatus,
    },
};
use alloc::{
    collections::{BTreeMap, VecDeque},
    sync::Arc,
};
use k210_pac::Interrupt;
use lazy_static::lazy_static;
use riscv::register::sie;

lazy_static! {
    pub static ref IRQMANAGER: Arc<SpinMutex<IrqManager>> =
        Arc::new(SpinMutex::new(IrqManager::new()));
}

pub struct IrqManager {
    pub plic_instance: Arc<dyn PlicDevice>,
    pub task_list: BTreeMap<usize, VecDeque<Arc<TaskControlBlock>>>,
}

impl IrqManager {
    pub fn new() -> Self {
        Self {
            plic_instance: PLIC_DRIVE.clone(),
            task_list: BTreeMap::new(),
        }
    }

    pub fn register_irq(&mut self, source: usize, hartid: usize) {
        self.plic_instance.enable(source, hartid);
        self.plic_instance.set_priority(1, source);
        self.task_list.insert(source, VecDeque::new());
    }
    pub fn inqueue(&mut self, irq: usize, task: Arc<TaskControlBlock>) {
        if let Some(queue) = self.task_list.get_mut(&irq) {
            queue.push_back(task)
        }
    }
    pub fn dequeue(&mut self, irq: usize) -> Option<Arc<TaskControlBlock>> {
        if let Some(queue) = self.task_list.get_mut(&irq) {
            queue.pop_front()
        } else {
            None
        }
    }

    pub fn set_thershold(&mut self, value: u32, hartid: usize) {
        self.plic_instance.set_thershold(value, hartid)
    }

    pub fn complete(&self, irq: usize, hartid: usize) {
        self.plic_instance.complete(irq, hartid)
    }
}

pub fn irq_init(hartid: usize) {
    unsafe {
        let mut irq_manager = IRQMANAGER.lock().unwrap();

        irq_manager.set_thershold(0, hartid);
        #[cfg(feature = "board_k210")]
        {
            sie::set_ssoft();
            irq_manager.register_irq(Interrupt::DMA0 as usize, hartid);
            irq_manager.register_irq(Interrupt::UARTHS as usize, hartid);
        }
        #[cfg(feature = "board_qemu")]
        {
            sie::set_sext();
            // irq_manager.register_irq(1, hartid);
            irq_manager.register_irq(10, hartid);
        }
    }
}

#[no_mangle]
pub fn wait_for_irq_and_run_next(irq: usize) {
    intr_check!();
    let mut irq_manager = IRQMANAGER.lock().unwrap();
    if let Some(task) = take_current_task() {
        let mut task_inner = task.inner_lock_access().unwrap();
        task_inner.task_status = TaskStatus::Waiting;
        let task_cx_ptr = &mut task_inner.task_cx as *mut TaskContext;
        drop(task_inner);
        irq_manager.inqueue(irq, task);
        drop(irq_manager);
        schedule(task_cx_ptr);
    } else {
        panic!("too early irq {}", irq);
    }
}

pub fn handler_ext() {
    intr_check!();
    let mut irq_manager = IRQMANAGER.lock().unwrap();
    let irq = PLIC_DRIVE.current(current_hartid());
    #[cfg(feature = "board_k210")]
    match irq {
        0 => {}
        27 => {
            BLOCK_DEVICE.handler_interrupt();
            match irq_manager.dequeue(irq) {
                Some(task) => add_task(task),
                None => {}
            }
        }
        33 => {
            UART_DEVICE.handler_interrupt();
            match irq_manager.dequeue(irq) {
                Some(task) => add_task(task),
                None => {}
            }
        }
        _ => {
            panic!("Unsupported irq {}", irq)
        }
    }
    #[cfg(feature = "board_qemu")]
    match irq {
        0 => {}
        27 => {
            BLOCK_DEVICE.handler_interrupt();
            match irq_manager.dequeue(irq) {
                Some(task) => add_task(task),
                None => {}
            }
        }
        10 => {
            UART_DEVICE.handler_interrupt();
            match irq_manager.dequeue(irq) {
                Some(task) => add_task(task),
                None => {}
            }
        }
        _ => {
            panic!("Unsupported irq {}", irq)
        }
    }
    if irq != 0 {
        irq_manager.complete(irq, current_hartid());
    }
}

impl Debug for IrqManager {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        f.write_str("irqmanager")
    }
}
