use buddy_system_allocator::FrameAllocator;
use riscv::register::sepc;

use crate::irq::IRQMANAGER;
use crate::mm::FRAME_ALLOCATOR;
use crate::sbi::{console_putchar, shutdown};
use crate::sync::intr_off;
use crate::task::{current_hartid, current_kstack_top, current_process, TASK_MANAGER};
use core::fmt::{self, Write};
use core::panic::PanicInfo;
use core::sync::atomic::{AtomicBool, Ordering};

struct Stderr;

impl Write for Stderr {
    fn write_str(&mut self, s: &str) -> core::fmt::Result {
        for ch in s.chars() {
            console_putchar(ch as usize)
        }
        Ok(())
    }
}

fn print(args: fmt::Arguments) {
    Stderr.write_fmt(args).unwrap();
}

#[macro_export]
macro_rules! debug_println {
    ($fmt: literal $(, $($arg: tt)+)?) => {
        $crate::lang_items::print(format_args!(concat!("[LotusOS] ",$fmt, "\r\n") $(, $($arg)+)?))
    }
}

#[macro_export]
macro_rules! debug_print {
    ($fmt: literal $(, $($arg: tt)+)?) => {
        $crate::lang_items::print(format_args!(concat!("",$fmt) $(, $($arg)+)?))
    }
}

static mut PANICLOCK: AtomicBool = AtomicBool::new(false);

// #[panic_handler]
// pub unsafe fn panic(info: &PanicInfo) -> ! {
//     // extern "C" {
//     //     fn jmp_panic() -> !;
//     // }
//     // jmp_panic()
//     panic_handler(info);
//     loop {
        
//     }
// }
#[no_mangle]
#[panic_handler]
pub unsafe fn panic_handler(info: &PanicInfo) -> ! {
    intr_off();
    while PANICLOCK
        .compare_exchange(false, true, Ordering::Acquire, Ordering::Relaxed)
        .is_err()
    {
        while PANICLOCK.load(Ordering::Relaxed) {}
    }
    debug_println!("Panic! Sepc {:#x}",sepc::read());
    let task_manager = TASK_MANAGER.force_lock();
    let irq_manager = IRQMANAGER.force_lock();
    let frame_alloc = FRAME_ALLOCATOR.force_lock();
    match info.location() {
        Some(location) => {
            debug_println!(
                "hartid {} panicked at [{}], {}:{}:{}$",
                current_hartid(),
                info.message().unwrap(),
                location.file(),
                location.line(),
                location.column()
            );
        }
        None => debug_println!(
            "hartid {} panicked at '{}'$",
            current_hartid(),
            info.message().unwrap()
        ),
    };

    debug_println!("Ready Tasks: {}", task_manager.ready_queue.len());
    for (irq, v) in irq_manager.task_list.iter() {
        debug_println!("Irq {} {}", irq, v.len())
    }
    debug_println!(
        "Total Page: {} {} {}",
        frame_alloc.current,
        frame_alloc.end,
        frame_alloc.recycled.len()
    );
    PANICLOCK.store(false, Ordering::Release);
    shutdown()
}

#[allow(unused)]
unsafe fn backtrace() {
    let mut fp: usize;
    let stop = current_kstack_top();
    core::arch::asm!("mv {}, s0", out(reg) fp);
    println!("---START BACKTRACE---");
    for i in 0..10 {
        if fp == stop {
            break;
        }
        println!("#{}:ra={:#x}", i, *((fp - 8) as *const usize));
        fp = *((fp - 16) as *const usize);
    }
    println!("---END   BACKTRACE---");
}
