use super::{frame_alloc, FrameTracker};
use super::{PTEFlags, PageTable, PageTableEntry};
use super::{PhysAddr, PhysPageNum, VirtAddr, VirtPageNum};
use super::{StepByOne, VPNRange};
use crate::config::{MEMORY_END, MMIO, PAGE_SIZE, TRAMPOLINE, USER_HEAP_SIZE};
use crate::fs::{open_file, vfs_read, OSInode, OpenFlags};
use crate::mm::memory_set;
use crate::sync::{SpinMutex, UPSafeCell};
use crate::syscall::MmapProt;
use crate::task::{Auxv, AuxvType, CLONE_FLAGS};
use alloc::collections::BTreeMap;
use alloc::string::{String, ToString};
use alloc::sync::Arc;
use alloc::vec::Vec;
use riscv::asm;
use riscv::paging::PTE;
use core::fmt::Debug;
use core::mem;
use elf_rs::{Elf, ElfFile, ElfHeader, ProgramHeaderFlags, ProgramType};
use k210_pac::aes::en;
use lazy_static::*;
use log::{warn, info};
use riscv::addr::Page;
use riscv::register::{satp, sepc};
use riscv::register::scause::Exception;
use riscv::register::scause::Trap;

extern "C" {
    fn stext();
    fn etext();
    fn srodata();
    fn erodata();
    fn sdata();
    fn edata();
    fn sbss_with_stack();
    fn ebss();
    fn ekernel();
    fn strampoline();
}

lazy_static! {
    pub static ref KERNEL_SPACE: Arc<SpinMutex<MemorySet>> =
        Arc::new(SpinMutex::new(MemorySet::new_kernel()));
}

pub fn kernel_token() -> usize {
    KERNEL_SPACE.lock().unwrap().token()
}

pub struct MemorySet {
    pub page_table: PageTable,
    pub areas: Vec<MapArea>,
}

impl MemorySet {
    pub fn new_bare() -> Self {
        Self {
            page_table: PageTable::new(),
            areas: Vec::new(),
        }
    }

    pub fn new_clone(pt: PageTable) -> Self {
        Self {
            page_table: pt,
            areas: Vec::new(),
        }
    }

    pub fn token(&self) -> usize {
        self.page_table.token()
    }

    pub fn mprotect(&mut self, addr: usize, len: usize, prot: MmapProt) {
        let start: VirtAddr = addr.into();
        let end: VirtAddr = (addr + len).into();
        self.areas
            .iter_mut()
            .find(|area| area.contains(addr, addr + len))
            .map(|area| {
                let mut perm = MapPermission::U;
                if prot.contains(MmapProt::PROT_READ) {
                    perm |= MapPermission::R;
                }
                if prot.contains(MmapProt::PROT_WRITE) {
                    perm |= MapPermission::W;
                }
                if prot.contains(MmapProt::PROT_EXEC) {
                    perm |= MapPermission::X;
                }
                area.map_perm = perm;
            });
    }

    /// Assume that no conflicts.
    pub fn insert_framed_area(
        &mut self,
        start_va: VirtAddr,
        end_va: VirtAddr,
        permission: MapPermission,
        map_type: MapType,
    ) {
        self.push(MapArea::new(start_va, end_va, map_type, permission), None);
    }

    pub fn insert_framed_area_unmap(
        &mut self,
        start_va: VirtAddr,
        end_va: VirtAddr,
        permission: MapPermission,
        map_type: MapType,
    ) {
        self.areas
            .push(MapArea::new(start_va, end_va, map_type, permission));
    }

    pub fn remove_area_with_start_vpn(&mut self, start_vpn: VirtPageNum) {
        if let Some((idx, area)) = self
            .areas
            .iter_mut()
            .enumerate()
            .find(|(_, area)| area.vpn_range.get_start() == start_vpn)
        {
            area.unmap(&mut self.page_table);
            self.areas.remove(idx);
        }
    }

    fn push(&mut self, mut map_area: MapArea, data: Option<&[u8]>) {
        map_area.map(&mut self.page_table);
        if let Some(data) = data {
            map_area.copy_data(&mut self.page_table, data);
        }
        self.areas.push(map_area);
    }

    fn push_fd(&mut self, mut map_area: MapArea, offset: usize, len: usize, inode: Arc<OSInode>) {
        map_area.map(&mut self.page_table);
        map_area.copy_data_fd(&mut self.page_table, offset, len as isize, inode);
        self.areas.push(map_area);
    }

    /// Mention that trampoline is not collected by areas.
    fn map_trampoline(&mut self) {
        self.page_table.map(
            VirtAddr::from(TRAMPOLINE).into(),
            PhysAddr::from(strampoline as usize).into(),
            PTEFlags::R | PTEFlags::X,
        );
    }

    /// Without kernel stacks.
    pub fn new_kernel() -> Self {
        let mut memory_set = Self::new_bare();
        // map trampoline
        memory_set.map_trampoline();
        // map kernel sections
        println!(".text [{:#x}, {:#x})", stext as usize, etext as usize);
        println!(".rodata [{:#x}, {:#x})", srodata as usize, erodata as usize);
        println!(".data [{:#x}, {:#x})", sdata as usize, edata as usize);
        println!(
            ".bss [{:#x}, {:#x})",
            sbss_with_stack as usize, ebss as usize
        );
        println!("mapping .text section");
        memory_set.push(
            MapArea::new(
                (stext as usize).into(),
                (etext as usize).into(),
                MapType::Identical,
                MapPermission::R | MapPermission::X,
            ),
            None,
        );
        println!("mapping .rodata section");
        memory_set.push(
            MapArea::new(
                (srodata as usize).into(),
                (erodata as usize).into(),
                MapType::Identical,
                MapPermission::R,
            ),
            None,
        );
        println!("mapping .data section");
        memory_set.push(
            MapArea::new(
                (sdata as usize).into(),
                (edata as usize).into(),
                MapType::Identical,
                MapPermission::R | MapPermission::W,
            ),
            None,
        );
        println!("mapping .bss section");
        memory_set.push(
            MapArea::new(
                (sbss_with_stack as usize).into(),
                (ebss as usize).into(),
                MapType::Identical,
                MapPermission::R | MapPermission::W,
            ),
            None,
        );
        println!("mapping physical memory");
        memory_set.push(
            MapArea::new(
                (ekernel as usize).into(),
                MEMORY_END.into(),
                MapType::Identical,
                MapPermission::R | MapPermission::W,
            ),
            None,
        );
        println!("mapping memory-mapped registers");
        for pair in MMIO {
            memory_set.push(
                MapArea::new(
                    (*pair).0.into(),
                    ((*pair).0 + (*pair).1).into(),
                    MapType::Identical,
                    MapPermission::R | MapPermission::W,
                ),
                None,
            );
        }
        memory_set
    }

    pub fn load_elf_phdrs<'a>(
        mut header: &'a mut Vec<u8>,
        elf: Elf,
        inode: Arc<OSInode>,
    ) -> Elf<'a> {
        let elf_header = elf.elf_header();
        let num = elf_header.program_header_entry_num() as usize;
        let size = elf_header.program_header_entry_size() as usize;
        let offset = elf_header.program_header_offset() as usize;
        header.resize(offset + size * num, 0);
        inode.seek(0);
        inode.read(&mut header);
        Elf::from_bytes(header.as_slice()).unwrap()
    }

    pub fn load_elf_interp(inter: Arc<OSInode>, elf: &Elf, memory_set: &mut Self) {
        let elf_header = elf.elf_header();
        let base = 0x5000_0000;
        for header in elf.program_header_iter() {
            if header.ph_type() == ProgramType::LOAD {
                let start_va: VirtAddr = (header.vaddr() as usize + base).into();
                let end_va: VirtAddr =
                    (header.vaddr() as usize + header.memsz() as usize + base).into();
                let mut map_perm = MapPermission::U;
                let flag = header.flags();
                if flag.contains(ProgramHeaderFlags::EXECUTE) {
                    map_perm |= MapPermission::X;
                }
                if flag.contains(ProgramHeaderFlags::READ) {
                    map_perm |= MapPermission::R;
                }
                if flag.contains(ProgramHeaderFlags::WRITE) {
                    map_perm |= MapPermission::W;
                }
                let map_area = MapArea::new(start_va, end_va, MapType::Framed, map_perm);

                memory_set.push_fd(
                    map_area,
                    header.offset() as usize,
                    header.filesz() as usize,
                    inter.clone(),
                );
            }
        }
    }

    pub fn from_elf_fd(inode: Arc<OSInode>) -> (Self, usize, usize, Vec<Auxv>) {
        let mut elf;
        let mut inter_elf = None;
        let mut elf_interpreter = None;
        let mut interpreter = None;
        let header_buf = &mut [0; 64];
        let mut header = Vec::new();
        let mut inter_header = Vec::new();
        inode.seek(0);
        assert_eq!(inode.read(header_buf), 64);
        elf = Elf::from_bytes(header_buf).unwrap();
        elf = Self::load_elf_phdrs(&mut header, elf, inode.clone());
        let elf_header = elf.elf_header();

        // 判断是否是为需要动态链接
        for header in elf.program_header_iter() {
            if header.ph_type() == ProgramType::INTERP {
                let mut buf: Vec<u8> = Vec::new();
                buf.resize(header.memsz() as usize, 0);
                vfs_read(
                    inode.clone(),
                    &mut buf,
                    header.offset() as usize,
                    header.memsz() as usize,
                );
                assert!(buf[buf.len() - 1] == 0);
                let ld = String::from_utf8(buf).unwrap();
                // /lib/ld-musl-riscv64-sf.so.1 仅仅是对 /libc.so 软连接，而fat32不支持，因此直接转换成libc.so
                elf_interpreter = Some(if ld.starts_with("/lib/ld-musl-riscv64-sf.so.1") {
                    "/libc.so".to_string()
                } else {
                    ld
                });
                // println!(
                //     "Find ProgramType::INTERP {}",
                //     elf_interpreter.as_ref().unwrap()
                // );
                interpreter = open_file(
                    elf_interpreter.as_ref().unwrap().as_str(),
                    OpenFlags::RDONLY,
                );
                break;
            }
        }

        // 如果为动态加载类型则读取其解释器的ELF头部
        if elf_interpreter.is_some() {
            assert!(interpreter.is_some());
            if let Some(inter) = interpreter.as_ref() {
                inter.seek(0);
                assert_eq!(inter.read(header_buf), 64);
                let elf = Elf::from_bytes(header_buf).unwrap();
                inter_elf = Some(Self::load_elf_phdrs(&mut inter_header, elf, inter.clone()));
            }
        }

        let mut max_end_vpn = VirtPageNum(0);
        let mut memory_set = Self::new_bare();
        // map trampoline
        memory_set.map_trampoline();
        for header in elf.program_header_iter() {
            if header.ph_type() == ProgramType::LOAD {
                let start_va: VirtAddr = (header.vaddr() as usize).into();
                let end_va: VirtAddr = (header.vaddr() as usize + header.memsz() as usize).into();
                let mut map_perm = MapPermission::U;
                let flag = header.flags();
                if flag.contains(ProgramHeaderFlags::EXECUTE) {
                    map_perm |= MapPermission::X;
                }
                if flag.contains(ProgramHeaderFlags::READ) {
                    map_perm |= MapPermission::R;
                }
                if flag.contains(ProgramHeaderFlags::WRITE) {
                    map_perm |= MapPermission::W;
                }
                let map_area = MapArea::new(start_va, end_va, MapType::Framed, map_perm);
                max_end_vpn = map_area.vpn_range.get_end();

                memory_set.push_fd(
                    map_area,
                    header.offset() as usize,
                    header.filesz() as usize,
                    inode.clone(),
                );
            }
        }

        if interpreter.is_some() && inter_elf.is_some() {
            // println!("load_elf_interp");
            Self::load_elf_interp(
                interpreter.as_ref().unwrap().clone(),
                inter_elf.as_ref().unwrap(),
                &mut memory_set,
            );
        }
        let mut vaddr = 0;
        for header in elf.program_header_iter() {
            if header.ph_type() == ProgramType::LOAD {
                vaddr = header.vaddr();
                break;
            }
        }

        let mut auxv: Vec<Auxv> = Vec::new();
        auxv.push(Auxv::new(
            AuxvType::AT_PHDR,
            vaddr as usize + elf_header.program_header_offset() as usize,
        ));
        auxv.push(Auxv::new(
            AuxvType::AT_PHENT,
            elf_header.program_header_entry_size() as usize,
        ));
        auxv.push(Auxv::new(
            AuxvType::AT_PHNUM,
            elf_header.program_header_entry_num() as usize,
        ));
        auxv.push(Auxv::new(AuxvType::AT_PAGESZ, PAGE_SIZE));
        auxv.push(Auxv::new(AuxvType::AT_BASE, 0x5000_0000));
        auxv.push(Auxv::new(AuxvType::AT_FLAGS, 0));
        auxv.push(Auxv::new(AuxvType::AT_ENTRY, elf.entry_point() as usize));
        auxv.push(Auxv::new(AuxvType::AT_UID, 0));
        auxv.push(Auxv::new(AuxvType::AT_EUID, 0));
        auxv.push(Auxv::new(AuxvType::AT_GID, 0));
        auxv.push(Auxv::new(AuxvType::AT_EGID, 0));
        auxv.push(Auxv::new(AuxvType::AT_HWCAP, 0));
        auxv.push(Auxv::new(AuxvType::AT_CLKTCK, 0x64));
        auxv.push(Auxv::new(AuxvType::AT_SECURE, 0));

        let program_break: VirtAddr = max_end_vpn.into();
        // map program break
        memory_set.insert_framed_area_unmap(
            program_break,
            (program_break.0 + USER_HEAP_SIZE).into(),
            MapPermission::U | MapPermission::W | MapPermission::R,
            MapType::UserHeap,
        );

        let program_break = program_break.into();

        if interpreter.is_some() {
            (
                memory_set,
                program_break,
                inter_elf.as_ref().unwrap().entry_point() as usize + 0x5000_0000,
                auxv,
            )
        } else {
            (memory_set, program_break, elf.entry_point() as usize, auxv)
        }
    }

    pub fn from_user_cow(parent_space: &mut MemorySet) -> Self {
        let mut memory_set = Self::new_bare();
        memory_set.map_trampoline();
        for area in parent_space.areas.iter() {
            let mut new_area = MapArea::from_another(area);
            let mut map_perm = PTEFlags::from_bits(area.map_perm.bits()).unwrap();
            for (vpn, frame) in area.data_frames.iter() {
                let (new_vpn, new_ppn) = match area.map_type {
                    MapType::Identical
                    | MapType::SignalStack
                    | MapType::MmapPrivate
                    | MapType::Framed
                    | MapType::UserStack
                    | MapType::UserHeap
                    | MapType::MmapNormal => {
                        // Copy On Write
                        map_perm.remove(PTEFlags::W);
                        new_area.data_frames.insert(*vpn, frame.clone());
                        (*vpn, frame.ppn)
                    }
                    MapType::TrapContext => {
                        // 必须单独设置否则无法正常进入内核态
                        let new_frame = frame_alloc().unwrap();
                        let new_ppn = new_frame.ppn;
                        new_frame
                            .ppn
                            .get_bytes_array()
                            .copy_from_slice(frame.ppn.get_bytes_array());
                        new_area.data_frames.insert(*vpn, Arc::new(new_frame));
                        (*vpn, new_ppn)
                    }
                    MapType::KernelStack => panic!("shoud not be here"),
                };
                memory_set.page_table.map(new_vpn, new_ppn, map_perm);
                parent_space.page_table.update_map(*vpn, frame.ppn, map_perm)
            }
            memory_set.areas.push(new_area);
        }
        memory_set
    }

    pub fn from_existed_user(user_space: &MemorySet) -> MemorySet {
        let mut memory_set = Self::new_bare();
        // map trampoline
        memory_set.map_trampoline();
        // copy data sections/trap_context/user_stack
        for area in user_space.areas.iter() {
            let new_area = MapArea::from_another(area);
            memory_set.push(new_area, None);
            // copy data from another space
            for vpn in area.vpn_range {
                let src_ppn = user_space.translate(vpn).unwrap().ppn();
                let dst_ppn = memory_set.translate(vpn).unwrap().ppn();
                dst_ppn
                    .get_bytes_array()
                    .copy_from_slice(src_ppn.get_bytes_array());
            }
        }
        memory_set
    }

    pub fn handler_pte_fault(memory_set: &mut MemorySet, stval: usize, fault: Trap) -> Result<(),()> {
        let vaddr: VirtAddr = stval.into();
        let vpn: VirtPageNum = vaddr.floor().into();
        let src_pte = if let Some(pte) = memory_set.translate(vpn) {
            pte
        } else {
            return Err(());
        };
        // println!("PageFault: {:#x} {:?} {:#x}",stval,fault,sepc::read());
        let src_ppn = src_pte.ppn();
        if let Some(area) = memory_set.areas.iter_mut().find(|area| {
            let start: VirtAddr = area.vpn_range.get_start().into();
            let end: VirtAddr = area.vpn_range.get_end().into();
            return start.0 <= stval && end.0 >= stval;
        }) {
            
            let frame = frame_alloc().unwrap();
            let dst_ppn = frame.ppn;
            let mut flags;
        
            if let Some(frame_tracker) = area.data_frames.get_mut(&vpn) {
                flags = src_pte.flags() | PTEFlags::W | PTEFlags::R | PTEFlags::X;
                memory_set.page_table.unmap(vpn);
                dst_ppn
                    .get_bytes_array()
                    .copy_from_slice(src_ppn.get_bytes_array());
                *frame_tracker = Arc::new(frame);
            } else {
                flags = PTEFlags::from_bits(area.map_perm.bits).unwrap();
                area.data_frames.insert(vpn, Arc::new(frame));
            }
            memory_set.page_table.map(vpn, dst_ppn, flags);
            let pte = memory_set.translate(vpn).unwrap();
            Ok(())
        } else {
            info!("PTE fault on unmapped area {:#x}",stval);
            Err(())
        }
    }

    pub fn activate(&self) {
        let satp = self.page_table.token();
        unsafe {
            satp::write(satp);
            core::arch::asm!("sfence.vma");
        }
    }
    pub fn translate(&self, vpn: VirtPageNum) -> Option<PageTableEntry> {
        self.page_table.translate(vpn)
    }
    pub fn recycle_data_pages(&mut self) {
        //*self = Self::new_bare();
        self.areas.clear();
    }
}

pub struct MapArea {
    pub vpn_range: VPNRange,
    data_frames: BTreeMap<VirtPageNum, Arc<FrameTracker>>,
    map_type: MapType,
    map_perm: MapPermission,
}

impl MapArea {
    pub fn new(
        start_va: VirtAddr,
        end_va: VirtAddr,
        map_type: MapType,
        map_perm: MapPermission,
    ) -> Self {
        let start_vpn: VirtPageNum = start_va.floor();
        let end_vpn: VirtPageNum = end_va.ceil();
        Self {
            vpn_range: VPNRange::new(start_vpn, end_vpn),
            data_frames: BTreeMap::new(),
            map_type,
            map_perm,
        }
    }

    pub fn contains(&self, start: usize, end: usize) -> bool {
        self.vpn_range.get_start().0 << 12 <= start && self.vpn_range.get_end().0 << 12 >= end
    }

    pub fn from_another(another: &MapArea) -> Self {
        Self {
            vpn_range: VPNRange::new(another.vpn_range.get_start(), another.vpn_range.get_end()),
            data_frames: BTreeMap::new(),
            map_type: another.map_type,
            map_perm: another.map_perm,
        }
    }

    // 首次分配内存时，需要分配一个新的frame，并且将新的frame映射到虚拟地址空间中
    pub fn map_one(&mut self, page_table: &mut PageTable, vpn: VirtPageNum) {
        let ppn: PhysPageNum;
        match self.map_type {
            MapType::Identical => {
                ppn = PhysPageNum(vpn.0);
            }
            MapType::Framed
            | MapType::SignalStack
            | MapType::TrapContext
            | MapType::UserStack
            | MapType::KernelStack
            | MapType::MmapNormal
            | MapType::MmapPrivate
            | MapType::UserHeap => {
                let frame = frame_alloc().unwrap();
                ppn = frame.ppn;
                self.data_frames.insert(vpn, Arc::new(frame));
            }
        }
        let pte_flags = PTEFlags::from_bits(self.map_perm.bits).unwrap();
        page_table.map(vpn, ppn, pte_flags);
    }

    pub fn unmap_one(&mut self, page_table: &mut PageTable, vpn: VirtPageNum) {
        self.data_frames.remove(&vpn);
        page_table.unmap(vpn);
    }

    pub fn map(&mut self, page_table: &mut PageTable) {
        for vpn in self.vpn_range {
            self.map_one(page_table, vpn);
        }
    }

    pub fn unmap(&mut self, page_table: &mut PageTable) {
        for vpn in self.vpn_range {
            self.unmap_one(page_table, vpn);
        }
    }

    /// data: start-aligned but maybe with shorter length
    /// assume that all frames were cleared before
    pub fn copy_data(&mut self, page_table: &mut PageTable, data: &[u8]) {
        assert_eq!(self.map_type, MapType::Framed);
        let mut start: usize = 0;
        let mut current_vpn = self.vpn_range.get_start();
        let len = data.len();
        loop {
            let src = &data[start..len.min(start + PAGE_SIZE)];
            let dst = &mut page_table
                .translate(current_vpn)
                .unwrap()
                .ppn()
                .get_bytes_array()[..src.len()];
            dst.copy_from_slice(src);
            start += PAGE_SIZE;
            if start >= len {
                break;
            }
            current_vpn.step();
        }
    }

    pub fn copy_data_fd(
        &mut self,
        page_table: &mut PageTable,
        mut offset: usize,
        mut len: isize,
        inode: Arc<OSInode>,
    ) {
        assert_eq!(self.map_type, MapType::Framed);
        let mut current_vpn = self.vpn_range.get_start();
        inode.seek(offset);
        let mut left: usize = offset % PAGE_SIZE;
        let mut right: usize = core::cmp::min(PAGE_SIZE, offset + len as usize);
        loop {
            let dst = &mut page_table
                .translate(current_vpn)
                .unwrap()
                .ppn()
                .get_bytes_array()[left..right];
            inode.read(dst);
            len -= dst.len() as isize;
            if len <= 0 {
                break;
            }
            left = 0;
            right = core::cmp::min(PAGE_SIZE, len as usize);
            current_vpn.step();
        }
    }
}

#[derive(Copy, Clone, PartialEq, Debug)]
pub enum MapType {
    Identical,
    Framed,
    TrapContext,
    UserStack,
    UserHeap,
    KernelStack,
    SignalStack,
    MmapNormal,
    MmapPrivate,
}

bitflags! {
    pub struct MapPermission: u8 {
        const R = 1 << 1;
        const W = 1 << 2;
        const X = 1 << 3;
        const U = 1 << 4;
    }
}

#[allow(unused)]
pub fn remap_test() {
    let mut kernel_space = KERNEL_SPACE.lock().unwrap();
    let mid_text: VirtAddr = ((stext as usize + etext as usize) / 2).into();
    let mid_rodata: VirtAddr = ((srodata as usize + erodata as usize) / 2).into();
    let mid_data: VirtAddr = ((sdata as usize + edata as usize) / 2).into();
    assert_eq!(
        kernel_space
            .page_table
            .translate(mid_text.floor())
            .unwrap()
            .writable(),
        false
    );
    assert_eq!(
        kernel_space
            .page_table
            .translate(mid_rodata.floor())
            .unwrap()
            .writable(),
        false,
    );
    assert_eq!(
        kernel_space
            .page_table
            .translate(mid_data.floor())
            .unwrap()
            .executable(),
        false,
    );
    println!("remap_test passed!");
}
