use core::mem::size_of;

use crate::{
    config::TRAMPOLINE,
    mm::{translated_byte_buffer, translated_bytes, UserBuffer},
    task::{current_process, current_task},
    trap::{trap_handler, trap_return, TrapContext},
};
use alloc::{boxed::Box, collections::BTreeMap, sync::Arc, vec::Vec};

bitflags! {
    pub struct SignalFlag: usize {
        const SA_NOCLDSTOP = 0x1;
        const SA_NOCLDWAIT = 0x2;
        const SA_SIGINFO   = 0x4;
        const SA_RESTART   = 0x10000000;
        const SA_NODEFER   = 0x40000000;
        const SA_RESETHAND = 0x80000000;
        const SA_RESTORER  = 0x04000000;
    }
}

#[derive(Debug)]
#[repr(C)]
pub struct SignalAction {
    handler: usize,
    flags: usize,
    restorer: usize,
    mask: [u32; 2],
}

pub struct SignalManager {
    pending_signals: Vec<isize>,
    sig_mask: [usize; 16],
    handlers: BTreeMap<isize, SignalAction>,
}

impl SignalManager {
    pub fn new() -> Self {
        Self {
            handlers: BTreeMap::new(),
            sig_mask: [0; 16],
            pending_signals: Vec::new(),
        }
    }
    pub fn add_handler(&mut self, signal: isize, action: SignalAction) {
        self.handlers.insert(signal, action);
    }

    pub fn send_signal(&mut self, signal: isize) {
        self.pending_signals.push(signal);
    }

    pub fn block(&mut self, set: &[usize; 16]) {
        for i in 0..16 {
            self.sig_mask[i] |= set[i];
        }
    }

    pub fn unblock(&mut self, set: &[usize; 16]) {
        for i in 0..16 {
            self.sig_mask[i] &= !set[i];
        }
    }

    pub fn setblocks(&mut self, set: &[usize; 16]) {
        for i in 0..16 {
            self.sig_mask[i] = set[i];
        }
    }

    pub fn read_olg(&self, old: &mut [usize; 16]) {
        for i in 0..16 {
            old[i] = self.sig_mask[i];
        }
    }
}
#[repr(C)]
#[derive(Default)]
pub struct SigInfo {
    si_signo: i32,
    si_errno: i32,
    si_code: i32,
    si_pid: i32,
    si_uid: u32,
    si_status: i32,
    si_utime: isize,
    si_stime: isize,
    si_value: isize,
    si_int: i32,
    si_ptr: usize,
    si_overrun: i32,
    si_timerid: i32,
    si_addr: usize,
    si_band: usize,
    si_fd: i32,
    si_addr_lsb: i16,
    si_lower: usize,
    si_upper: usize,
    si_pkey: i32,
    si_call_addr: usize,
    si_syscall: i32,
    si_arch: u32,
}
#[repr(C)]
#[derive(Default)]
pub struct SignalStack {
    ss_sp: usize,
    ss_flags: i32,
    ss_size: usize,
}
#[repr(C)]
pub struct MContext {
    __grep: [usize; 64],
    __fcsr: u32,
    __reserved: [u32; 3],
}

impl Default for MContext {
    fn default() -> Self {
        Self {
            __grep: [0; 64],
            __fcsr: 0,
            __reserved: [0; 3],
        }
    }
}

#[repr(C)]
#[derive(Default)]
pub struct UContext {
    uc_flag: usize,
    uc_link: usize,
    uc_stack: SignalStack,
    uc_sigmask: [usize; 16],
    uc_mcontext: MContext,
}

pub fn handler_exist_signal() {
    extern "C" {
        fn __alltraps();
        fn __signal_trampoline();
    }
    let process = current_process();
    let process_inner = process.inner_lock_access().unwrap();
    let token = process_inner.get_user_token();
    let mut signal_manager = process_inner.signal_manager.lock().unwrap();
    let task = current_task().unwrap();
    let task_inner = task.inner_lock_access().unwrap();
    if let Some(signal) = signal_manager.pending_signals.pop() {
        if let Some(handler) = signal_manager.handlers.get(&signal) {
            task_inner.backup_trap_cx();
            let trap_cx = task_inner.get_trap_cx();
            let mut signal_stack = task_inner.res.as_ref().unwrap().get_signal_stack();
            let flags = SignalFlag::from_bits(handler.flags).unwrap();

            trap_cx.sepc = handler.handler;
            trap_cx.x[1] = handler.restorer;
            trap_cx.x[10] = signal as usize;
            if flags.contains(SignalFlag::SA_SIGINFO) {
                let siginfo = SigInfo::default();
                signal_stack -= size_of::<SigInfo>();
                // let mut buffer = UserBuffer::new(translated_byte_buffer(token, signal_stack as *const u8, core::mem::size_of::<SigInfo>()));
                // buffer.write(translated_bytes(&siginfo, size_of::<SigInfo>()));
                trap_cx.x[11] = signal_stack;
                let ucontext = UContext::default();
                signal_stack -= core::mem::size_of::<UContext>();
                // let mut buffer = UserBuffer::new(translated_byte_buffer(token, signal_stack as *const u8, core::mem::size_of::<SigInfo>()));
                // buffer.write(translated_bytes(&ucontext, size_of::<UContext>()));
                trap_cx.x[12] = signal_stack;
            }
            signal_stack -= signal_stack % 16;
            trap_cx.set_sp(signal_stack);
        }
    }
}
