#[allow(unused)]
use core::{
    cell::UnsafeCell,
    ops::{Deref, DerefMut},
    sync::atomic::{AtomicBool, Ordering},
};
use core::{fmt::Debug, sync::atomic::AtomicIsize};
use riscv::register::sstatus;

use crate::task::current_hartid;

pub struct SpinMutex<T: ?Sized> {
    pub(crate) lock: AtomicBool,
    pub(crate) cpu: UnsafeCell<isize>,
    data: UnsafeCell<T>,
}

unsafe impl<T> Sync for SpinMutex<T> {}
unsafe impl<T> Send for SpinMutex<T> {}

pub struct SpinMutexGuard<'a, T: ?Sized> {
    lock: &'a AtomicBool,
    cpu: &'a mut isize,
    data: &'a mut T,
}

impl<T> SpinMutex<T> {
    #[inline(always)]
    pub const fn new(user_data: T) -> SpinMutex<T> {
        SpinMutex {
            lock: AtomicBool::new(false),
            cpu: UnsafeCell::new(-1),
            data: UnsafeCell::new(user_data),
        }
    }
}

impl<T: ?Sized> SpinMutex<T> {
    #[inline(always)]
    pub fn is_locked(&self) -> bool {
        self.lock.load(Ordering::Relaxed)
    }

    #[inline(always)]
    pub fn lock(&self) -> Result<SpinMutexGuard<T>, ()> {
        intr_off();
        while self
            .lock
            .compare_exchange(false, true, Ordering::Acquire, Ordering::Relaxed)
            .is_err()
        {
            while self.is_locked() {}
        }
        unsafe { *self.cpu.get() = current_hartid() as isize }
        Ok(SpinMutexGuard {
            lock: &self.lock,
            cpu: unsafe { &mut *self.cpu.get() },
            data: unsafe { &mut *self.data.get() },
        })
    }

    #[inline(always)]
    pub fn try_lock(&self) -> Result<SpinMutexGuard<T>, ()> {
        intr_off();
        if self
            .lock
            .compare_exchange(false, true, Ordering::Acquire, Ordering::Relaxed)
            .is_ok()
        {
            unsafe { *self.cpu.get() = current_hartid() as isize }
            Ok(SpinMutexGuard {
                lock: &self.lock,
                cpu: unsafe { &mut *self.cpu.get() },
                data: unsafe { &mut *self.data.get() },
            })
        } else {
            Err(())
        }
    }

    #[inline(always)]
    pub fn force_lock(&self) -> SpinMutexGuard<T> {
        self.lock.store(true, Ordering::Release);
        SpinMutexGuard {
            lock: &self.lock,
            cpu: unsafe { &mut *self.cpu.get() },
            data: unsafe { &mut *self.data.get() },
        }
    }
}

impl<'a, T: ?Sized> Drop for SpinMutexGuard<'a, T> {
    fn drop(&mut self) {
        let cpu = *self.cpu;
        let current = current_hartid();
        if cpu != current as isize {
            panic!("release wrong cpu {} {}", cpu, current);
        }
        *self.cpu = -1;
        self.lock.store(false, Ordering::Release);
    }
}

impl<'a, T: ?Sized> Deref for SpinMutexGuard<'a, T> {
    type Target = T;
    fn deref(&self) -> &T {
        self.data
    }
}

impl<'a, T: ?Sized> DerefMut for SpinMutexGuard<'a, T> {
    fn deref_mut(&mut self) -> &mut T {
        self.data
    }
}

#[inline]
pub fn intr_on() {
    unsafe {
        sstatus::set_sie();
    }
}

#[inline]
pub fn intr_off() {
    unsafe {
        sstatus::clear_sie();
    }
}
