use core::cell::{BorrowMutError, RefCell, RefMut};

/// Wrap a static data structure inside it so that we are
/// able to access it without any `unsafe`.
///
/// We should only use it in uniprocessor.
///
/// In order to get mutable reference of inner data, call
/// `exclusive_access`.
pub struct UPSafeCell<T> {
    /// inner data
    pub inner: RefCell<T>,
}

unsafe impl<T> Sync for UPSafeCell<T> {}

impl<T> UPSafeCell<T> {
    /// User is responsible to guarantee that inner struct is only used in
    /// uniprocessor.
    pub unsafe fn new(value: T) -> Self {
        Self {
            inner: RefCell::new(value),
        }
    }
    /// Panic if the data has been borrowed.
    pub fn exclusive_access(&self) -> RefMut<'_, T> {
        self.inner.borrow_mut()
    }
    // 在调用 try_exclusive_access 的地方 unwrap 方便暴露 错误位置
    pub fn try_exclusive_access(&self) -> Result<RefMut<'_, T>, BorrowMutError> {
        match self.inner.try_borrow_mut() {
            Ok(item) => Ok(item),
            Err(err) => Err(err),
        }
    }
}

impl<T> Drop for UPSafeCell<T> {
    fn drop(&mut self) {
        drop(self);
    }
}
