use super::process::sys_brk;
use crate::drivers::BLOCK_DEVICE;
use crate::fatfs::file::Inode;
use crate::fatfs::io::SeekFrom;
use crate::fatfs::root_dir;
use crate::fs::ZeroDev;
use crate::fs::make_pipe;
use crate::fs::root;
use crate::fs::Dirent;
use crate::fs::File;
use crate::fs::Kstat;
use crate::fs::NullDev;
use crate::fs::OSInode;
use crate::fs::OpenFlags;
use crate::fs::VDirent;
use crate::mm::translated_bytes;
use crate::mm::translated_copy_ref;
use crate::mm::translated_ref;
use crate::mm::{translated_byte_buffer, translated_refmut, translated_str, UserBuffer};
use crate::sync::SpinMutex;
use crate::sync::UPSafeCell;
use crate::task::current_task;
use crate::task::{current_process, current_user_token};
use crate::timer::TimeSpec;
use crate::timer::TimeVal;
use alloc::collections::BTreeMap;
use alloc::string::String;
use alloc::string::ToString;
use alloc::sync::Arc;
use alloc::vec;
use core::mem::size_of;
use log::info;
use riscv::register::sepc;

const AT_FDCWD: isize = -100;

lazy_static::lazy_static!(
    static ref MOUNTFILES: SpinMutex<BTreeMap<String, Arc<dyn File>>> = SpinMutex::new(BTreeMap::new());
);

pub fn sys_write(fd: usize, buf: *const u8, len: usize) -> isize {
    let token = current_user_token();
    let process = current_process();
    let inner = process.inner_lock_access().unwrap();
    let fd_table = inner.fd_table.lock().unwrap();
    if !fd_table.tables.contains_key(&fd) {
        return -1;
    }
    if let Some(file) = &fd_table.get(fd).unwrap() {
        if !file.writable() {
            return -1;
        }
        let file = file.clone();
        drop(fd_table);
        drop(inner);
        file.write(UserBuffer::new(translated_byte_buffer(token, buf, len))) as isize
    } else {
        -1
    }
}

pub fn sys_read(fd: usize, buf: *const u8, len: usize, off: isize) -> isize {
    let token = current_user_token();
    let process = current_process();
    let inner = process.inner_lock_access().unwrap();
    let fd_table = inner.fd_table.lock().unwrap();
    if !fd_table.tables.contains_key(&fd) {
        return -1;
    }

    if let Some(file) = &fd_table.get(fd).unwrap() {
        let file = file.clone();
        if !file.readable() {
            return -1;
        }
        // release current task TCB manually to avoid multi-borrow
        drop(fd_table);
        drop(inner);
        if off >= 0 {
            file.seek(SeekFrom::Start(off as u64));
        }
        let a = file.read(UserBuffer::new(translated_byte_buffer(token, buf, len))) as isize;
        a
    } else {
        -1
    }
}

pub fn sys_open(fd: isize, path: *const u8, flags: i32) -> isize {
    let token = current_user_token();
    let process = current_process();
    let mut inner = process.inner_lock_access().unwrap();
    let mut fd_table = inner.fd_table.lock().unwrap();
    let path = translated_str(token, path).replace("./", "");
    // println!("open: {}", path);
    let dir = if fd >= 0 {
        let fd = fd as usize;
        if let Some(dir) = fd_table.get(fd).unwrap() {
            dir
        } else {
            return -1;
        }
    } else {
        &inner.dir_entry.as_ref().unwrap()
    };

    if path == ".".to_string() {
        let tmp = dir.clone();
        drop(dir);
        let fd = fd_table.alloc_fd();
        fd_table.tables.insert(fd, Some(tmp));
        return fd as isize;
    }

    let flag = OpenFlags::from_bits(flags as u32).unwrap();

    if path == "/dev/zero".to_string() {
        let fd = fd_table.alloc_fd();
        fd_table.tables.insert(fd, Some(Arc::new(ZeroDev{})));
        return fd as isize;
    }

    let (readable, writable) = flag.read_write();
    let file = if flag.contains(OpenFlags::CREATE) {
        dir.create(&path, readable, writable, false)
    } else {
        let directory = flag.contains(OpenFlags::DIRECTORY);
        dir.open(&path, readable, writable, directory)
    };
    if let Some(file) = file {
        let fd = fd_table.alloc_fd();
        fd_table.tables.insert(fd, Some(file));
        // println!("Open Ok");
        fd as isize
    } else {
        // println!("Open Error");
        -24
    }
}

pub fn sys_unlink(dirfd: isize, path: *const u8, _flags: usize) -> isize {
    let token = current_user_token();
    let process = current_process();
    let inner = process.inner_lock_access().unwrap();
    if dirfd < 0 {
        let path = translated_str(token, path).replace("./", "");
        if inner.dir_entry.as_ref().unwrap().remove(&path) {
            return 0;
        }
    }
    1
}

pub fn sys_mkdir(dirfd: isize, path: *const u8, _mode: usize) -> isize {
    let token = current_user_token();
    let process = current_process();
    let mut inner = process.inner_lock_access().unwrap();
    let path = translated_str(token, path).replace("./", "");
    let mut fd_table = inner.fd_table.lock().unwrap();
    let dir = if dirfd >= 0 {
        if let Some(dir) = fd_table.get(dirfd as usize).unwrap() {
            dir
        } else {
            return -1;
        }
    } else {
        &inner.dir_entry.as_ref().unwrap()
    };
    if let Some(file) = dir.create(&path, false, false, true) {
        let fd = fd_table.alloc_fd();
        fd_table.tables.insert(fd, Some(file));
        fd as isize
    } else {
        -1
    }
}

pub fn sys_chdir(path: *const u8) -> isize {
    let token = current_user_token();
    let process = current_process();
    let mut inner = process.inner_lock_access().unwrap();
    let path = translated_str(token, path).replace("./", "");
    if path == "/".to_string() {
        inner.dir_entry = Some(root());
        return 0;
    }
    let inode = inner
        .dir_entry
        .as_ref()
        .unwrap()
        .open(&path, true, true, true);
    match inode {
        Some(file) => {
            inner.dir_entry = Some(file);
            0
        }
        None => 1,
    }
}

pub fn sys_close(fd: usize) -> isize {
    let process = current_process();
    let mut inner = process.inner_lock_access().unwrap();
    let mut fd_table = inner.fd_table.lock().unwrap();
    if !fd_table.tables.contains_key(&fd) {
        return -1;
    }
    if let Some(inode) = fd_table.tables.remove(&fd) {
        inode.unwrap().close();
    }
    0
}

pub fn sys_pipe(pipe: *mut i32) -> isize {
    let process = current_process();
    let token = current_user_token();
    let mut inner = process.inner_lock_access().unwrap();
    let (pipe_read, pipe_write) = make_pipe();
    let read_fd = inner.alloc_fd();
    let mut fd_table = inner.fd_table.lock().unwrap();
    fd_table
        .tables
        .get_mut(&read_fd)
        .unwrap()
        .replace(pipe_read);
    let write_fd = fd_table.alloc_fd();
    fd_table
        .tables
        .get_mut(&write_fd)
        .unwrap()
        .replace(pipe_write);
    *translated_refmut(token, pipe) = read_fd as i32;
    *translated_refmut(token, unsafe { pipe.add(1) }) = write_fd as i32;
    0
}

pub fn sys_dup(fd: usize) -> isize {
    let token = current_user_token();
    let process = current_process();
    let mut process_inner = process.inner_lock_access().unwrap();
    let mut fd_table = process_inner.fd_table.lock().unwrap();
    let task = current_task().unwrap();
    let mut task_inner = task.inner_lock_access().unwrap();
    if fd_table.tables[&fd].is_none() {
        return -1;
    }
    let new_fd = fd_table.alloc_fd();
    if new_fd == usize::MAX {
        return -24;
    }
    let value = Arc::clone(fd_table.tables[&fd].as_ref().unwrap());
    fd_table.tables.get_mut(&new_fd).unwrap().replace(value);

    new_fd as isize
}

pub fn sys_dup3(old: usize, new: usize) -> isize {
    let process = current_process();
    let mut inner = process.inner_lock_access().unwrap();
    let mut fd_table = inner.fd_table.lock().unwrap();
    if fd_table.tables[&old].is_none() {
        return -1;
    }
    if fd_table.tables.contains_key(&new) {
        fd_table.tables.remove(&new).unwrap();
    }
    let value = Arc::clone(fd_table.tables[&old].as_ref().unwrap());
    fd_table.tables.insert(new, Some(value));
    new as isize
}

pub fn sys_getcwd(buf: *const u8, len: usize) -> isize {
    let token = current_user_token();
    let process = current_process();
    let inner = process.inner_lock_access().unwrap();
    let name = inner.dir_entry.as_ref().unwrap().name();
    let dir = name.as_bytes();
    for b in UserBuffer::new(translated_byte_buffer(token, buf, len)).buffers {
        b[0..dir.len()].copy_from_slice(dir);
    }
    1
}

pub fn sys_fstat(fd: isize, ptr: *const u8) -> isize {
    let token = current_user_token();
    let mut user_buf = UserBuffer::new(translated_byte_buffer(
        token,
        ptr,
        core::mem::size_of::<Kstat>(),
    ));
    let process = current_process();
    let inner = process.inner_lock_access().unwrap();
    if let Some(opt) = inner.fd_table.lock().unwrap().get(fd as usize) {
        if let Some(file) = opt {
            let mut stat = Kstat::default();
            file.kstat(&mut stat);
            user_buf.write(translated_bytes(&stat, size_of::<Kstat>()));
            return 0;
        }
    }
    1
}

pub fn sys_getdents(fd: isize, ptr: *mut Dirent, _len: usize) -> isize {
    let token = current_user_token();
    let process = current_process();
    let inner = process.inner_lock_access().unwrap();
    let dirent = translated_refmut(token, ptr);
    let ret = match inner.fd_table.lock().unwrap().get(fd as usize) {
        Some(Some(file)) => file.getdents(dirent),
        Some(None) => -1,
        None => -1,
    };
    ret
}

bitflags! {
    pub struct MmapProt: usize {
        const PROT_NONE = 0x0;
        const PROT_READ = 0x1;
        const PROT_WRITE = 0x2;
        const PROT_EXEC = 0x4;
    }
    pub struct MmapFlag: usize {
        const MAP_SHARED = 0x01;
        const MAP_PRIVATE = 0x02;
        const MAP_FIXED = 0x10;
        const MAP_ANONYMOUS = 0x20;
        const MAP_GROWSDOWN = 0x0100;
        const MAP_DENYWRITE = 0x0800;
        const MAP_EXECUTABLE = 0x1000;
        const MAP_LOCKED = 0x2000;
        const MAP_NORESERVE = 0x4000;
        const MAP_POPULATE = 0x8000;
        const MAP_NONBLOCK = 0x10000;
        const MAP_STACK = 0x20000;
    }
}

pub fn sys_mmap(addr: usize, len: usize, prot: usize, flag: usize, fd: isize, off: usize) -> isize {
    let task = current_task().unwrap();
    let mut task_inner = task.inner_lock_access().unwrap();
    let context = task_inner.get_trap_cx();
    let prot = MmapProt::from_bits(prot).unwrap();
    let flag = MmapFlag::from_bits(flag).unwrap();
    let res = task_inner.res.as_mut().unwrap();
    let addr = res.mmap(addr, len, prot, flag, fd, off);
    addr
}

pub fn sys_munmap(start: usize, _len: usize) -> isize {
    0
}

lazy_static::lazy_static!(
    static ref MOUNT :SpinMutex<BTreeMap<String, String>> = SpinMutex::new(BTreeMap::new());
);

pub fn sys_mount(device: *const u8, mntpoint: *const u8) -> isize {
    let token = current_user_token();
    let device = translated_str(token, device);
    let mntpoint = translated_str(token, mntpoint);
    let mut mount = MOUNT.lock().unwrap();
    mount.insert(mntpoint, device);
    0
}
pub fn sys_umount(mntpoint: *const u8) -> isize {
    let token = current_user_token();
    let mntpoint = translated_str(token, mntpoint);
    let mut mount = MOUNT.lock().unwrap();
    if mount.contains_key(&mntpoint) {
        mount.remove(&mntpoint);
    }
    0
}

#[repr(C)]
pub struct Iovec {
    iov_base: usize,
    iov_len: u32,
}

pub fn sys_writev(fd: usize, mut iovp: usize, iovcnt: usize) -> isize {
    let token = current_user_token();
    let process = current_process();
    let inner = process.inner_lock_access().unwrap();
    let ret = match inner.fd_table.lock().unwrap().get(fd) {
        Some(Some(file)) => {
            let mut len: isize = 0;
            for _ in 0..iovcnt {
                let iov = translated_ref(token, iovp as *const Iovec);
                let user_buf = UserBuffer::new(translated_byte_buffer(
                    token,
                    iov.iov_base as *const u8,
                    iov.iov_len as usize,
                ));
                file.write(user_buf);
                len += iov.iov_len as isize;
                iovp += size_of::<Iovec>();
            }
            len
        }
        _ => -1,
    };
    ret
}

pub fn sys_readv(fd: usize, mut iovp: usize, iovcnt: usize) -> isize {
    let token = current_user_token();
    let process = current_process();
    let inner = process.inner_lock_access().unwrap();
    let ret = match inner.fd_table.lock().unwrap().get(fd) {
        Some(Some(file)) => {
            let mut len: isize = 0;
            for _ in 0..iovcnt {
                let iov = translated_ref(token, iovp as *const Iovec);

                let user_buf = UserBuffer::new(translated_byte_buffer(
                    token,
                    iov.iov_base as *const u8,
                    iov.iov_len as usize,
                ));
                let n = file.read(user_buf);
                len += n as isize;
                if n < iov.iov_len as usize || n == 0 {
                    break;
                }
                iovp += size_of::<Iovec>();
            }
            len
        }
        _ => -1,
    };
    ret
}

pub fn sys_fstatat(fd: isize, pathname: *const u8, statbuf: *const u8, flags: isize) -> isize {
    let process = current_process();
    let inner = process.inner_lock_access().unwrap();
    let mut stat = Kstat::default();
    let mut user_buf = UserBuffer::new(translated_byte_buffer(
        inner.get_user_token(),
        statbuf,
        size_of::<Kstat>(),
    ));
    let path = translated_str(inner.get_user_token(), pathname);

    if path == ".".to_string() {
        inner.dir_entry.as_ref().unwrap().kstat(&mut stat);
        user_buf.write(stat.as_bytes());
        return 0;
    }
    if fd == -100 {
        if path.starts_with("/dev/null") {
            NullDev {}.kstat(&mut stat);
            user_buf.write(stat.as_bytes());
            return 0;
        }
    }
    if let Some(opt) = inner.fd_table.lock().unwrap().get(fd as usize) {
        opt.as_ref().unwrap().kstat(&mut stat);
    } else {
        return -1;
    }
    user_buf.write(stat.as_bytes());
    0
}

pub fn sys_lseek(fd: isize, off: isize, whence: isize) -> isize {
    let process = current_process();
    let inner = process.inner_lock_access().unwrap();
    let ret = if let Some(opt) = inner.fd_table.lock().unwrap().get(fd as usize) {
        let inode = opt.as_ref().unwrap();
        inode.seek(match whence {
            0 => SeekFrom::Start(off as u64),
            1 => SeekFrom::Current(off as i64),
            2 => SeekFrom::End(off as i64),
            _ => return -1,
        }) as isize
    } else {
        -1
    };
    ret
}

pub fn sys_fcntl(fd: isize, cmd: isize) -> isize {

    1
}

pub fn sys_utimensat(fd: isize, path: *const u8, time: *const u8, flag: isize) -> isize {
    let token = current_user_token();
    let process = current_process();
    let process_inner = process.inner_lock_access().unwrap();
    match fd {
        AT_FDCWD => {
            let path = translated_str(token, path);
        }
        -1 => {}
        _ => {
            let fd_table = process_inner.fd_table.lock().unwrap();
            let inode = fd_table.get(fd as usize).unwrap().as_ref().unwrap();
            unsafe {
                let atime = translated_ref(token, time as *const TimeSpec);
                let ctime =
                    translated_ref(token, time.add(size_of::<TimeSpec>()) as *const TimeSpec);
                inode.utimens(atime, ctime);
            }
        }
    }

    0
}
#[repr(C)]
struct Statvfs {
    f_bsize: usize,
    f_frsize: usize,
    f_block: usize,
    f_bfree: usize,
    f_bavail: usize,
    f_files: usize,
    f_ffree: usize,
    f_favail: usize,
    f_fsid: usize,
    f_flag: usize,
    f_namemax: usize,
    __reserved: [i32; 6],
}

pub fn sys_statvfs(fd: isize, buf: usize) -> isize {
    let mut statvfs = Statvfs {
        f_bsize: 4096,
        f_frsize: 4096,
        f_block: 100,
        f_bfree: 100,
        f_bavail: 0,
        f_files: 100,
        f_ffree: 0,
        f_favail: 0,
        f_fsid: 100,
        f_flag: 100,
        f_namemax: 255,
        __reserved: [0; 6],
    };
    let mut user_buf = UserBuffer::new(translated_byte_buffer(
        current_user_token(),
        buf as *const u8,
        size_of::<Statvfs>(),
    ));
    user_buf.write(translated_bytes(&statvfs, size_of::<Statvfs>()));
    0
}
