use core::mem::size_of;

use log::warn;

use crate::{
    mm::{translated_byte_buffer, translated_bytes, translated_ref, translated_refmut, UserBuffer},
    task::{current_process, current_user_token},
    timer::{self, TimeVal},
};
const SYSNAME: &[u8; 7] = b"LotusOs";
const NODENAME: &[u8; 7] = b"lotusOS";
const RELEASE: &[u8; 18] = b"2020/2/14 23:13:35";
const VERSION: &[u8; 6] = b"V0.0.1";
const MACHINE: &[u8; 9] = b"riscv64gc";
const DOMAINNAME: &[u8; 9] = b"localhost";
#[repr(C)]
pub struct Utsname {
    sysname: [u8; 65],
    nodename: [u8; 65],
    release: [u8; 65],
    version: [u8; 65],
    machine: [u8; 65],
    domainname: [u8; 65],
}

pub fn sys_uname(ptr: *const u8) -> isize {
    let token = current_user_token();
    let mut uname_buf = UserBuffer::new(translated_byte_buffer(token, ptr, size_of::<Utsname>()));
    let mut uname = Utsname {
        sysname: [0; 65],
        nodename: [0; 65],
        release: [0; 65],
        version: [0; 65],
        machine: [0; 65],
        domainname: [0; 65],
    };
    uname.sysname[0..SYSNAME.len()].copy_from_slice(SYSNAME);
    uname.nodename[0..NODENAME.len()].copy_from_slice(NODENAME);
    uname.release[0..RELEASE.len()].copy_from_slice(RELEASE);
    uname.machine[0..MACHINE.len()].copy_from_slice(MACHINE);
    uname.version[0..VERSION.len()].copy_from_slice(VERSION);
    uname.domainname[0..DOMAINNAME.len()].copy_from_slice(DOMAINNAME);
    uname_buf.write(translated_bytes(&uname, size_of::<Utsname>()));
    0
}

pub fn sys_get_time(ptr: *mut TimeVal) -> isize {
    let token = current_user_token();
    let mut time = translated_refmut(token, ptr);
    timer::set_timeval(time);
    0
}

pub fn sys_clk_gettime(clk_id: usize, timespec: *const u8) -> isize {
    let token = current_user_token();
    let mut time = UserBuffer::new(translated_byte_buffer(
        token,
        timespec,
        size_of::<TimeVal>(),
    ));
    let mut t = TimeVal { sec: 0, usec: 0 };
    timer::set_timeval(&mut t);
    unsafe {
        let slice = core::slice::from_raw_parts(&t as *const _ as *const u8, size_of::<TimeVal>());
        time.write(slice);
    }
    0
}

#[repr(C)]
pub struct SysInfo {
    uptime: usize,
    loads: [usize; 3],
    toalram: usize,
    freeram: usize,
    sharedram: usize,
    bufferram: usize,
    totalswap: usize,
    freeswap: usize,
    procs: u16,
    pad: u16,
    totalhigh: usize,
    freehigt: usize,
    mem_unit: u32,
    _f: [u8; 256],
}
pub fn sys_sysinfo(info: *const u8) -> isize {
    let token = current_user_token();
    0
}

static LOG: &str = "Hello LotusOS!\r\n";

pub fn sys_log(r#type: isize, ptr: *const u8, len: isize) -> isize {
    match r#type {
        3 => {
            let token = current_user_token();
            let mut user_buf = UserBuffer::new(translated_byte_buffer(token, ptr, len as usize));
            user_buf.write(LOG.as_bytes());
            return LOG.len() as isize;
        }
        10 => return LOG.len() as isize,
        _ => {
            panic!("Unknow log flag")
        }
    }
    0
}
#[repr(C)]
struct Timex {
    modes: i32,
    offset: usize,
    freq: usize,
    maxerror: usize,
    esterror: usize,
    status: i32,
    constant: usize,
    precision: usize,
    tolerance: usize,
    tick: usize,
    ppsfreq: usize,
    jitter: usize,
    shift: i32,
    stabil: usize,
    jitcnt: usize,
    calcnt: usize,
    errcnt: usize,
    tai: i32,
}

pub fn sys_adjtimex(timex: *const u8) -> isize {
    0
}

pub fn sys_getuid() -> isize {
    0
}

pub fn sys_geteuid() -> isize {
    0
}

pub fn sys_getgid() -> isize {
    0
}

pub fn sys_getegid() -> isize {
    0
}

pub fn sys_getresuid() -> isize {
    0
}

pub fn sys_ioctl(_fd: usize, _request: usize) -> isize {
    0
}
#[repr(C)]
pub struct Rlimit {
    rlim_cur: usize,
    rlim_max: usize,
}

pub fn sys_prlimit(pid: usize, resource: isize, new_limit: usize, old_limit: usize) -> isize {
    let token = current_user_token();
    let process = current_process();
    let mut inner = process.inner_lock_access().unwrap();
    let mut rlimit = Rlimit {
        rlim_cur: 0,
        rlim_max: 0,
    };

    match (new_limit, old_limit) {
        (0, 0) => {}
        (0, _) => {

            match resource {
                7 => {
                    let fd_table = inner.fd_table.lock().unwrap();
                    rlimit.rlim_cur = fd_table.fd_max;
                    rlimit.rlim_max = fd_table.fd_max;
                }
                _ => {
                    return 0;
                }
            }
            let mut user_buf = UserBuffer::new(translated_byte_buffer(
                token,
                old_limit as *const u8,
                size_of::<Rlimit>(),
            ));
            user_buf.write(translated_bytes(&rlimit, size_of::<Rlimit>()));
        }
        (_, 0) => {

            match resource {
                7 => {
                    let mut fd_table = inner.fd_table.lock().unwrap();
                    let limit = translated_ref(token, new_limit as *const Rlimit);
                    fd_table.fd_max = limit.rlim_max
                }
                _ => {
                    return 0;
                }
            }
        }
        _ => {}
    }
    0
}
