use crate::mm::{
    translated_byte_buffer, translated_bytes, translated_copy, translated_copy_ref,
    translated_copy_refmut, translated_ref, UserBuffer,
};
use crate::sync::{MutexBlocking, MutexSpin, Semaphore, SignalAction, SpinMutex};
use crate::syscall::{process, MmapProt};
use crate::task::{
    add_task, block_current_and_run_next, current_process, current_task, current_user_token,
    schedule, suspend_current_and_run_next, take_current_task, ProcessControlBlock, TaskContext,
    TaskControlBlock, TaskStatus,
};
use crate::timer::{add_timer, get_time_ms, TimeVal};
use alloc::sync::Arc;
use alloc::vec::Vec;
use core::mem::{size_of, transmute_copy};
use core::sync::atomic::{fence, Ordering};
use riscv::register::sepc;

pub fn sys_sleep(ms: *const TimeVal) -> isize {
    let token = current_user_token();
    let time = translated_ref(token, ms);
    let expire_ms = get_time_ms() + time.sec * 1000;
    let task = current_task().unwrap();
    add_timer(expire_ms, task);
    block_current_and_run_next();
    0
}

pub fn sys_mutex_create(blocking: bool) -> isize {
    let process = current_process();
    let mut process_inner = process.try_inner_exclusive_access().unwrap();
    if let Some(id) = process_inner
        .mutex_list
        .iter()
        .enumerate()
        .find(|(_, item)| item.is_none())
        .map(|(id, _)| id)
    {
        process_inner.mutex_list[id] = if !blocking {
            Some(Arc::new(MutexSpin::new()))
        } else {
            Some(Arc::new(MutexBlocking::new()))
        };
        id as isize
    } else {
        process_inner
            .mutex_list
            .push(Some(Arc::new(MutexSpin::new())));
        process_inner.mutex_list.len() as isize - 1
    }
}

pub fn sys_mutex_lock(mutex_id: usize) -> isize {
    let process = current_process();
    let process_inner = process.try_inner_exclusive_access().unwrap();
    let mutex = Arc::clone(process_inner.mutex_list[mutex_id].as_ref().unwrap());
    drop(process_inner);
    drop(process);
    mutex.lock();
    0
}

pub fn sys_mutex_unlock(mutex_id: usize) -> isize {
    let process = current_process();
    let process_inner = process.try_inner_exclusive_access().unwrap();
    let mutex = Arc::clone(process_inner.mutex_list[mutex_id].as_ref().unwrap());
    drop(process_inner);
    drop(process);
    mutex.unlock();
    0
}

pub fn sys_semaphore_creare(res_count: usize) -> isize {
    let process = current_process();
    let mut process_inner = process.try_inner_exclusive_access().unwrap();
    let id = if let Some(id) = process_inner
        .semaphore_list
        .iter()
        .enumerate()
        .find(|(_, item)| item.is_none())
        .map(|(id, _)| id)
    {
        process_inner.semaphore_list[id] = Some(Arc::new(Semaphore::new(res_count)));
        id
    } else {
        process_inner
            .semaphore_list
            .push(Some(Arc::new(Semaphore::new(res_count))));
        process_inner.semaphore_list.len() - 1
    };
    id as isize
}

pub fn sys_semaphore_up(sem_id: usize) -> isize {
    let process = current_process();
    let process_inner = process.try_inner_exclusive_access().unwrap();
    let sem = Arc::clone(process_inner.semaphore_list[sem_id].as_ref().unwrap());
    drop(process_inner);
    sem.up();
    0
}

pub fn sys_semaphore_down(sem_id: usize) -> isize {
    let process = current_process();
    let process_inner = process.try_inner_exclusive_access().unwrap();
    let sem = Arc::clone(process_inner.semaphore_list[sem_id].as_ref().unwrap());
    drop(process_inner);
    sem.down();
    0
}

const SIG_BLOCK: isize = 0;
const SIG_UNBLOCK: isize = 1;
const SIG_SETMASK: isize = 2;
pub fn sys_sigprocmask(how: isize, set: usize, old_set: usize) -> isize {
    let process = current_process();
    let process_inner = process.inner_lock_access().unwrap();
    let token = process_inner.get_user_token();
    let mut signal_manager = process_inner.signal_manager.lock().unwrap();
    if old_set != 0 {
        let mut tmp_set = [0usize; 16];
        signal_manager.read_olg(&mut tmp_set);
        let mut user_buf = UserBuffer::new(translated_byte_buffer(
            token,
            old_set as *const u8,
            16 * size_of::<usize>(),
        ));
        user_buf.write(translated_bytes(&tmp_set, 16 * size_of::<usize>()));
    }
    if set == 0 {
        return 0;
    }
    let mut sig_set: [usize; 16] = translated_copy(token, set, 16 * size_of::<usize>());
    match how {
        SIG_BLOCK => signal_manager.block(&sig_set),
        SIG_UNBLOCK => signal_manager.unblock(&sig_set),
        SIG_SETMASK => signal_manager.setblocks(&sig_set),
        _ => return -1,
    }
    0
}

pub fn sys_sigtimedwait() -> isize {
    0
}

const FUTEX_WAIT: usize = 0;
const FUTEX_WAKE: usize = 1;
const FUTEX_FD: usize = 2;
const FUTEX_REQUEUE: usize = 3;
const FUTEX_CMP_REQUEUE: usize = 4;
const FUTEX_WAKE_OP: usize = 5;
const FUTEX_LOCK_PI: usize = 6;
const FUTEX_UNLOCK_PI: usize = 7;
const FUTEX_TRYLOCK_PI: usize = 8;
const FUTEX_WAIT_BITSET: usize = 9;

pub struct WaitTask {
    pub task: Arc<TaskControlBlock>,
    pub uaddr: usize,
    val: i32,
}

lazy_static::lazy_static!(
    pub static ref WAIT_TASK_LIST:SpinMutex<Vec<WaitTask>>  = SpinMutex::new(Vec::new());
);

pub fn sys_futex(
    uaddr: usize,
    futex_op: usize,
    val: i32,
    val2: usize,
    uaddr2: usize,
    val3: i32,
) -> isize {
    let token = current_user_token();
    let op = futex_op & !0x80;
    // println!("Uaddr {:#x} Uaddr2 {:#x} Op {:#x}",uaddr, uaddr2, op);
    let value = translated_ref(token, uaddr as *const i32);
    match op {
        FUTEX_WAIT => {
            if val == *value {
                let mut wait_list = WAIT_TASK_LIST.lock().unwrap();
                let task = take_current_task().unwrap();
                let mut task_inner = task.inner_lock_access().unwrap();
                let task_cx_ptr = &mut task_inner.task_cx as *mut TaskContext;
                task_inner.task_status = TaskStatus::Blocking;
                drop(task_inner);
                wait_list.push(WaitTask { task, uaddr, val });
                drop(wait_list);
                schedule(task_cx_ptr);
                0
            } else {
                11
            }
        }
        FUTEX_WAKE => {
            let mut n = 0;
            let mut wait_list = WAIT_TASK_LIST.lock().unwrap();
            let idx = wait_list
                .iter()
                .enumerate()
                .find(|(_, v)| v.uaddr == uaddr)
                .map(|(idx, _)| idx);
            if let Some(idx) = idx {
                let wait_task = wait_list.remove(idx);
                add_task(wait_task.task);
            }
            n
        }
        FUTEX_REQUEUE => {
            let mut wait_list = WAIT_TASK_LIST.lock().unwrap();
            for i in 0..val {
                let idx = wait_list
                    .iter()
                    .enumerate()
                    .find(|(_, v)| v.uaddr == uaddr)
                    .map(|(idx, _)| idx);
                if let Some(idx) = idx {
                    let wait_task = wait_list.remove(idx);
                    add_task(wait_task.task);
                }
            }
            for task in wait_list.iter_mut() {
                if uaddr == task.uaddr {
                    task.uaddr = uaddr2;
                }
            }
            0
        }
        _ => panic!("Not Impl {}", op),
    }
}

pub fn sys_tkill(tid: isize, signal: isize) -> isize {
    let process = current_process();
    let mut process_inner = process.inner_lock_access().unwrap();
    let thread = process_inner.get_thread(tid as usize);
    let thread_inner = thread.inner_lock_access().unwrap();
    let mut signal_manager = thread_inner.signal_manager.lock().unwrap();
    signal_manager.send_signal(signal);
    0
}

pub fn sys_get_robust_list(pid: usize, head_ptr: usize, len_ptr: usize) -> isize {
    0
}

pub fn sys_set_robust_list(head: usize, len: usize) -> isize {
    0
}

pub fn sys_sigaction(sig: isize, act: usize, old: usize) -> isize {
    let token = current_user_token();
    let sigaction: SignalAction = translated_copy(token, act, size_of::<SignalAction>());
    let process = current_process();
    let process_inner = process.inner_lock_access().unwrap();
    let mut signal_manager = process_inner.signal_manager.lock().unwrap();
    signal_manager.add_handler(sig, sigaction);
    0
}

pub fn sys_sigreturn() -> isize {
    let task = current_task().unwrap();
    let task_inner = task.inner_lock_access().unwrap();
    task_inner.restore_trap_cx();
    task_inner.get_trap_cx().x[10] as isize
}

const MEMBARRIER_CMD_QUERY: isize = 0;
const MEMBARRIER_CMD_REGISTER_PRIVATE_EXPEDITED: isize = 1 << 4;

pub fn sys_membarrier(cmd: isize, flags: usize, cpu_id: isize) -> isize {
    match cmd {
        MEMBARRIER_CMD_QUERY => 0,
        MEMBARRIER_CMD_REGISTER_PRIVATE_EXPEDITED => {
            fence(Ordering::Acquire);
            0
        }
        _ => -1,
    }
}

pub fn sys_mprotect(addr: usize, len: usize, prot: usize) -> isize {
    let process = current_process();
    let process_inner = process.inner_lock_access().unwrap();
    let mut memory_set = process_inner.memory_set.lock().unwrap();
    let prot = MmapProt::from_bits(prot).unwrap();

    memory_set.mprotect(addr, len, prot);
    0
}
