use super::ProcessControlBlock;
use crate::config::{
    KERNEL_STACK_SIZE, PAGE_SIZE, SIGNAL_STACK_SIZE, TRAMPOLINE, TRAP_CONTEXT_BASE, USER_MMAP_BASE,
    USER_STACK_BASE, USER_STACK_SIZE, USER_STACK_UMMAP,
};
use crate::fatfs::io::SeekFrom;
use crate::fs::OSInode;
use crate::mm::{
    translated_byte_buffer, MapPermission, MapType, PhysPageNum, UserBuffer, VirtAddr, KERNEL_SPACE,
};
use crate::sync::SpinMutex;
use crate::syscall::{MmapFlag, MmapProt};
use crate::task::current_process;
use alloc::collections::VecDeque;
use alloc::{
    sync::{Arc, Weak},
    vec::Vec,
};
use lazy_static::*;
#[derive(Debug)]
pub struct RecycleAllocator {
    current: usize,
    recycled: Vec<usize>,
}

impl RecycleAllocator {
    pub fn new(current: usize) -> Self {
        RecycleAllocator {
            current,
            recycled: Vec::new(),
        }
    }
    pub fn alloc(&mut self) -> usize {
        if let Some(id) = self.recycled.pop() {
            id
        } else {
            self.current += 1;
            self.current - 1
        }
    }
    pub fn dealloc(&mut self, id: usize) {
        assert!(id < self.current);
        assert!(
            self.recycled.iter().find(|i| **i == id).is_none(),
            "id {} has been deallocated!",
            id
        );
        self.recycled.push(id);
    }
}

lazy_static! {
    static ref PID_ALLOCATOR: SpinMutex<RecycleAllocator> =
        SpinMutex::new(RecycleAllocator::new(1));
    static ref KSTACK_ALLOCATOR: SpinMutex<RecycleAllocator> =
        SpinMutex::new(RecycleAllocator::new(0));
}
#[derive(Debug)]
pub struct PidHandle(pub usize);

pub type TidHandle = PidHandle;

pub fn tid_alloc() -> TidHandle {
    // PidHandle(PID_ALLOCATOR.lock().alloc())
    let mut pid = PID_ALLOCATOR.lock().unwrap();
    let pid = pid.alloc();
    PidHandle(pid)
}

pub fn pid_alloc() -> PidHandle {
    // PidHandle(PID_ALLOCATOR.lock().alloc())
    let mut pid = PID_ALLOCATOR.lock().unwrap();
    let pid = pid.alloc();
    PidHandle(pid)
}

impl Drop for PidHandle {
    fn drop(&mut self) {
        PID_ALLOCATOR.lock().unwrap().dealloc(self.0);
    }
}

/// Return (bottom, top) of a kernel stack in kernel space.
pub fn kernel_stack_position(kstack_id: usize) -> (usize, usize) {
    let top = TRAMPOLINE - kstack_id * (KERNEL_STACK_SIZE + PAGE_SIZE);
    let bottom = top - KERNEL_STACK_SIZE;
    (bottom, top)
}

pub struct KernelStack(pub usize);

pub fn kstack_alloc() -> KernelStack {

    let kstack_id = KSTACK_ALLOCATOR.lock().unwrap().alloc();
    let (kstack_bottom, kstack_top) = kernel_stack_position(kstack_id);
    KERNEL_SPACE.lock().unwrap().insert_framed_area(
        kstack_bottom.into(),
        kstack_top.into(),
        MapPermission::R | MapPermission::W,
        MapType::KernelStack,
    );
    KernelStack(kstack_id)
}

impl Drop for KernelStack {
    fn drop(&mut self) {
        let (kernel_stack_bottom, _) = kernel_stack_position(self.0);
        let kernel_stack_bottom_va: VirtAddr = kernel_stack_bottom.into();
        KERNEL_SPACE
            .lock()
            .unwrap()
            .remove_area_with_start_vpn(kernel_stack_bottom_va.into());
    }
}

impl KernelStack {
    #[allow(unused)]
    pub fn push_on_top<T>(&self, value: T) -> *mut T
    where
        T: Sized,
    {
        let kernel_stack_top = self.get_top();
        let ptr_mut = (kernel_stack_top - core::mem::size_of::<T>()) as *mut T;
        unsafe {
            *ptr_mut = value;
        }
        ptr_mut
    }
    pub fn get_top(&self) -> usize {
        let (_, kernel_stack_top) = kernel_stack_position(self.0);
        kernel_stack_top
    }
}

pub struct TaskUserRes {
    pub res_id: usize,
    pub program_break: usize,
    pub process: Weak<ProcessControlBlock>,
    pub mmap_set: VecDeque<MmapArea>,
    pub mmap_end: usize,
}

fn trap_cx_bottom_from_tid(tid: usize) -> usize {
    TRAP_CONTEXT_BASE - tid * PAGE_SIZE
}

fn ustack_bottom_from_tid(tid: usize) -> usize {
    USER_STACK_BASE + tid * (PAGE_SIZE + USER_STACK_SIZE) + USER_STACK_UMMAP
}

// tid + 1 为了当tid为0时与user_stack_bottom区分开
fn signal_stack_bottom_from_tid(tid: usize) -> usize {
    USER_STACK_BASE - (tid + 1) * (PAGE_SIZE + SIGNAL_STACK_SIZE)
}

impl TaskUserRes {
    pub fn new(
        process: Arc<ProcessControlBlock>,
        alloc_user_res: bool,
        program_break: usize,
    ) -> Self {
        let res_id = process.inner_lock_access().unwrap().alloc_res_id();
        let task_user_res = Self {
            res_id,
            program_break,
            process: Arc::downgrade(&process),
            mmap_set: VecDeque::new(),
            mmap_end: USER_MMAP_BASE,
        };
        if alloc_user_res {
            task_user_res.alloc_user_res();
        }
        task_user_res
    }

    pub fn alloc_user_res(&self) {
        let process = self.process.upgrade().unwrap();
        let mut process_inner = process.inner_lock_access().unwrap();

        // alloc user stack
        let ustack_bottom = ustack_bottom_from_tid(self.res_id);
        let ustack_top = ustack_bottom + USER_STACK_SIZE + 1;
        let mut memory_set = process_inner.memory_set.lock().unwrap();
        memory_set.insert_framed_area(
            ustack_bottom.into(),
            ustack_top.into(),
            MapPermission::R | MapPermission::W | MapPermission::U,
            MapType::UserStack,
        );

        // alloc unmap user stack
        memory_set.insert_framed_area_unmap(
            (ustack_bottom - USER_STACK_UMMAP).into(),
            ustack_bottom.into(),
            MapPermission::R | MapPermission::W | MapPermission::U,
            MapType::UserStack,
        );

        // alloc trap_cx
        let trap_cx_bottom = trap_cx_bottom_from_tid(self.res_id);
        let trap_cx_top = trap_cx_bottom + PAGE_SIZE;
        memory_set.insert_framed_area(
            trap_cx_bottom.into(),
            trap_cx_top.into(),
            MapPermission::R | MapPermission::W,
            MapType::TrapContext,
        );

        //alloc signal stack
        let signal_stack_bottom = signal_stack_bottom_from_tid(self.res_id);
        let signal_stack_top = signal_stack_bottom + SIGNAL_STACK_SIZE;

        memory_set.insert_framed_area(
            signal_stack_bottom.into(),
            signal_stack_top.into(),
            MapPermission::R | MapPermission::W | MapPermission::U,
            MapType::SignalStack,
        );
    }

    fn dealloc_user_res(&self) {
        // dealloc tid
        let process = self.process.upgrade().unwrap();
        let mut process_inner = process.inner_lock_access().unwrap();
        // dealloc ustack manually
        let ustack_bottom_va: VirtAddr = ustack_bottom_from_tid(self.res_id).into();
        let mut memory_set = process_inner.memory_set.lock().unwrap();
        memory_set.remove_area_with_start_vpn(ustack_bottom_va.into());
        // dealloc trap_cx manually
        let trap_cx_bottom_va: VirtAddr = trap_cx_bottom_from_tid(self.res_id).into();
        memory_set.remove_area_with_start_vpn(trap_cx_bottom_va.into());

        // dealloc signal_stack manually
        let signal_stack_bottom_va: VirtAddr = signal_stack_bottom_from_tid(self.res_id).into();
        memory_set.remove_area_with_start_vpn(signal_stack_bottom_va.into());
    }

    pub fn get_signal_stack(&self) -> usize {
        let signal_stack_bottom = signal_stack_bottom_from_tid(self.res_id);
        signal_stack_bottom + SIGNAL_STACK_SIZE
    }

    pub fn dealloc_res_id(&self) {
        let process = self.process.upgrade().unwrap();
        let mut process_inner = process.inner_lock_access().unwrap();
        process_inner.dealloc_res_id(self.res_id);
    }

    pub fn trap_cx_user_va(&self) -> usize {
        trap_cx_bottom_from_tid(self.res_id)
    }

    pub fn trap_cx_ppn(&self) -> PhysPageNum {
        let process = self.process.upgrade().unwrap();
        let mut process_inner = process.inner_lock_access().unwrap();
        let trap_cx_bottom_va: VirtAddr = trap_cx_bottom_from_tid(self.res_id).into();
        let ppn = process_inner
            .memory_set
            .lock()
            .unwrap()
            .translate(trap_cx_bottom_va.into())
            .unwrap()
            .ppn();
        ppn
    }

    pub fn ustack_base(&self) -> usize {
        USER_STACK_BASE
    }

    pub fn ustack_top(&self) -> usize {
        ustack_bottom_from_tid(self.res_id) + USER_STACK_SIZE
    }

    pub fn brk(&mut self, offset: usize) -> usize {
        match offset {
            0 => self.program_break,
            _ => {
                self.program_break = offset;
                self.program_break
            }
        }
    }

    pub fn mmap(
        &mut self,
        start: usize,
        len: usize,
        prot: MmapProt,
        flag: MmapFlag,
        fd: isize,
        off: usize,
    ) -> isize {
        let process = current_process();
        let mut inner = process.inner_lock_access().unwrap();
        let page_len = PAGE_SIZE * (len / PAGE_SIZE);
        let (mmap_start, mmap_end) = if start == 0 {
            let addr = self.mmap_end;
            self.mmap_end += page_len;
            (addr, addr + page_len)
        } else {
            (start, start + page_len)
        };

        let mut memory_set = inner.memory_set.lock().unwrap();

        let mut perm = MapPermission::U;

        if prot.contains(MmapProt::PROT_READ) {
            perm |= MapPermission::R;
        }
        if prot.contains(MmapProt::PROT_WRITE) {
            perm |= MapPermission::W;
        }
        if prot.contains(MmapProt::PROT_EXEC) {
            perm |= MapPermission::X;
        }

        let contains_vpn = memory_set
            .areas
            .iter()
            .find(|area| area.contains(mmap_start, mmap_end))
            .map(|area| area.vpn_range.get_start());
        if contains_vpn.is_some() {

            // memory_set.remove_area_with_start_vpn(contains_vpn.unwrap());
        } else {
            memory_set.insert_framed_area(
                mmap_start.into(),
                mmap_end.into(),
                perm,
                MapType::MmapNormal,
            );
        }

        let user_buf = UserBuffer::new(translated_byte_buffer(
            memory_set.token(),
            mmap_start as *const u8,
            len,
        ));
        if fd == -1 {
            for ch in user_buf {
                unsafe {
                    *ch = 0;
                }
            }
            return mmap_start as isize;
        }

        let ret = if let Some(mut opt) = inner.fd_table.lock().unwrap().get(fd as usize) {
            let inode = opt.as_ref().unwrap().clone();

            inode.seek(SeekFrom::Start(off as u64));
            let n = inode.read(user_buf);
            mmap_start as isize
        } else {
            -1
        };
        ret
    }
}

pub struct MmapArea {
    start_addr: usize,
    len: usize,
    port: usize,
    flag: usize,
    fd: Arc<OSInode>,
    fd_offset: usize,
}

impl Drop for TaskUserRes {
    fn drop(&mut self) {
        self.dealloc_res_id();
        self.dealloc_user_res();
    }
}
