mod context;
mod id;
mod manager;
mod process;
mod processor;
mod switch;
mod task;
use crate::{
    fs::{open_file, OpenFlags, OSInode},
    mm::translated_refmut,
    syscall::WAIT_TASK_LIST, fatfs::{root_dir, file::{Inode, ArrFile}}, sync::UPSafeCell,
};
use alloc::sync::Arc;
use core::mem::size_of;
use lazy_static::*;
use manager::fetch_task;

pub use context::TaskContext;
pub use id::{kstack_alloc, pid_alloc, KernelStack, PidHandle};
pub use manager::{add_task, TASK_MANAGER};
pub use process::{
    do_fork, Auxv, AuxvType, FdTable, ProcessControlBlock, ProcessControlBlockInner, CLONE_FLAGS,
};
pub use processor::{
    current_hartid, current_kstack_top, current_process, current_processor, current_task,
    current_trap_cx, current_trap_cx_user_va, current_user_token, enter_kernel, exit_kernel,
    init_hart, run_tasks, schedule, take_current_task,
};
use switch::__switch;
pub use task::{TaskControlBlock, TaskStatus, Tms};
static mut UNUSED: [u8; size_of::<TaskContext>()] = [0; size_of::<TaskContext>()];

pub fn suspend_current_and_run_next() {
    intr_check!();
    // There must be an application running.
    let task = current_task().unwrap();
    // ---- access current TCB exclusively
    let mut task_inner = task.inner_lock_access().unwrap();
    let task_cx_ptr = &mut task_inner.task_cx as *mut TaskContext;
    // Change status to Ready
    task_inner.task_status = TaskStatus::Ready;
    drop(task_inner);
    // ---- release current TCB
    // push back to ready queue.
    // add_task(task);
    // jump to scheduling cycle
    schedule(task_cx_ptr);
}

pub fn block_current_and_run_next() {
    let task = take_current_task().unwrap();
    let mut task_inner = task.inner_lock_access().unwrap();
    let task_cx_ptr = &mut task_inner.task_cx as *mut TaskContext;
    task_inner.task_status = TaskStatus::Blocking;
    drop(task_inner);
    schedule(task_cx_ptr);
}

pub fn exit_current_and_run_next(exit_code: i32) {
    let task = take_current_task().unwrap();
    let mut task_inner = task.inner_lock_access().unwrap();
    let task_flags = task_inner.flags;
    let process = task.process.upgrade().unwrap();

    let tid = process.tid.0;
    let pid = process.pid.0;
    // 记录进程返回值
    task_inner.exit_code = Some(exit_code);
    task_inner.res = None;
    // here we do not remove the thread since we are still using the kstack
    // it will be deallocated when sys_waittid is called
    drop(task_inner);
    drop(task);
    // however, if this is the main thread of current process
    // the process should terminate at once
    if !task_flags.contains(CLONE_FLAGS::CLONE_THREAD) {
        let mut process_inner = process.inner_lock_access().unwrap();
        // mark this process as a zombie process
        process_inner.is_zombie = true;
        // record exit code of main process
        process_inner.exit_code = exit_code;

        if process_inner.children.len() > 0 {
            let mut initproc_inner = INITPROC.inner_lock_access().unwrap();
            for child in process_inner.children.iter() {
                child.try_inner_exclusive_access().unwrap().parent =
                    Some(Arc::downgrade(&INITPROC));
                initproc_inner.children.push(child.clone());
            }
        }
        // move all child processes under init process

        // deallocate user res (including tid/trap_cx/ustack) of all threads
        // it has to be done before we dealloc the whole memory_set
        // otherwise they will be deallocated twice
        for task in process_inner.tasks.iter().filter(|t| t.is_some()) {
            let task = task.as_ref().unwrap();
            let mut task_inner = task.inner_lock_access().unwrap();
            task_inner.res = None;
        }

        process_inner.children.clear();
        // deallocate other data in user space i.e. program code/data section
        process_inner
            .memory_set
            .lock()
            .unwrap()
            .recycle_data_pages();
    } else {
        let mut process_inner = process.inner_lock_access().unwrap();
        let token = process_inner.get_user_token();
        if process_inner.ctid != 0 {
            *translated_refmut(token, process_inner.ctid as *mut i32) = 0;
        }
        let mut wait_list = WAIT_TASK_LIST.lock().unwrap();

        let idx = wait_list
            .iter()
            .enumerate()
            .find(|(_, v)| v.uaddr == process_inner.ctid)
            .map(|(idx, _)| idx);
        if let Some(idx) = idx {
            let wait_task = wait_list.remove(idx);
            add_task(wait_task.task);

        }
    }

    drop(process);
    // we do not have to save task context
    // let mut _unused = TaskContext::zero_init();
    unsafe {
        let mut _unused: TaskContext = core::mem::transmute_copy(&UNUSED);
        schedule(&mut _unused as *mut _);
    }
}

lazy_static! {
    pub static ref INITPROC: Arc<ProcessControlBlock> = {
        let v = include_bytes!("../usertests");
        let inode = Arc::new(unsafe{ UPSafeCell::new(Inode::Arr(ArrFile{
            v: v.to_vec(),
            offset: 0
        }))});
        let file = Arc::new(OSInode::new(true, true, inode));

        // let initproc = "usertests";
        // let file = open_file(initproc, OpenFlags::RDONLY).unwrap();

        ProcessControlBlock::new(file)
    };
}

pub fn add_initproc() {
    let _initproc = INITPROC.clone();
}
