use core::cell::{BorrowMutError, RefMut};
use core::clone;
use core::fmt::Debug;

use super::id::{RecycleAllocator, TidHandle};
use super::{add_task, current_process};
use super::{current_user_token, TaskControlBlock};
use super::{pid_alloc, PidHandle};
use crate::fs::{root, File, OSInode, Stdin, Stdout};
use crate::mm::{kernel_token, translated_refmut, MemorySet, KERNEL_SPACE};
use crate::sync::{Mutex, Semaphore, SignalManager, SpinMutex, SpinMutexGuard, UPSafeCell};
use crate::task::current_task;
use crate::task::id::tid_alloc;
use crate::trap::{trap_handler, TrapContext};
use alloc::collections::BTreeMap;
use alloc::string::String;
use alloc::sync::{Arc, Weak};
use alloc::vec::Vec;
use log::info;
use riscv::register::sepc;

bitflags! {
    pub struct CLONE_FLAGS:usize {
        const NONE = 0x0;
        const CSIGNAL = 0x000000ff;
        const CLONE_VM = 0x00000100;
        const CLONE_FS = 0x00000200;
        const CLONE_FILES = 0x00000400;
        const CLONE_SIGHAND = 0x00000800;
        const CLONE_PIDFD = 0x00001000;
        const CLONE_PTRACE = 0x00002000;
        const CLONE_VFORK = 0x00004000;
        const CLONE_PARENT = 0x00008000;
        const CLONE_THREAD = 0x00010000;
        const CLONE_NEWNS = 0x00020000;
        const CLONE_SYSVSEM = 0x00040000;
        const CLONE_SETTLS = 0x00080000;
        const CLONE_PARENT_SETTID = 0x00100000;
        const CLONE_CHILD_CLEARTID = 0x00200000;
        const CLONE_DETACHED = 0x00400000;
        const CLONE_UNTRACED = 0x00800000;
        const CLONE_CHILD_SETTID = 0x01000000;
        const CLONE_NEWCGROUP = 0x02000000;
        const CLONE_NEWUTS = 0x04000000;
        const CLONE_NEWIPC = 0x08000000;
        const CLONE_NEWUSER = 0x10000000;
        const CLONE_NEWPID = 0x20000000;
        const CLONE_NEWNET = 0x40000000;
        const CLONE_IO = 0x80000000;
    }
}

#[allow(unused)]
#[allow(non_camel_case_types)]
#[derive(Clone, Copy)]
pub enum AuxvType {
    AT_NULL = 0,
    AT_IGNORE,
    AT_EXECFD,
    AT_PHDR,
    AT_PHENT,
    AT_PHNUM,
    AT_PAGESZ,
    AT_BASE,
    AT_FLAGS,
    AT_ENTRY,
    AT_NOTELF,
    AT_UID,
    AT_EUID,
    AT_GID,
    AT_EGID,
    AT_PLATFORM,
    AT_HWCAP,
    AT_CLKTCK,
    AT_SECURE = 23,
    AT_BASE_PLATFORM = 24,
    AT_RANDOM = 25,
    AT_HWCAP2 = 26,
    AT_EXECFN = 31,
}

pub struct Auxv {
    a_type: AuxvType,
    val: usize,
}

impl Auxv {
    pub fn new(a_type: AuxvType, val: usize) -> Self {
        Self { a_type, val }
    }
}

pub struct ProcessControlBlock {
    // immutable
    pub pid: Arc<PidHandle>,
    pub tid: Arc<TidHandle>,
    // mutable
    pub inner: SpinMutex<ProcessControlBlockInner>,
}

pub struct ProcessControlBlockInner {
    pub ctid: usize,
    pub is_zombie: bool,
    pub exit_code: i32,
    pub parent: Option<Weak<ProcessControlBlock>>,

    // resources
    pub tasks: Vec<Option<Arc<TaskControlBlock>>>,
    pub children: Vec<Arc<ProcessControlBlock>>,
    pub thread_group: Vec<Arc<ProcessControlBlock>>,
    pub signal_manager: Arc<SpinMutex<SignalManager>>,
    pub memory_set: Arc<SpinMutex<MemorySet>>,
    pub fd_table: Arc<SpinMutex<FdTable>>,
    pub task_res_allocator: Arc<SpinMutex<RecycleAllocator>>,
    pub mutex_list: Vec<Option<Arc<dyn Mutex>>>,
    pub semaphore_list: Vec<Option<Arc<Semaphore>>>,
    pub dir_entry: Option<Arc<dyn File + Send + Sync>>,
}

impl ProcessControlBlockInner {
    #[allow(unused)]
    pub fn get_user_token(&self) -> usize {
        self.memory_set.lock().unwrap().token()
    }

    pub fn alloc_fd(&mut self) -> usize {
        self.fd_table.lock().unwrap().alloc_fd()
    }

    pub fn alloc_res_id(&mut self) -> usize {
        self.task_res_allocator.lock().unwrap().alloc()
    }

    pub fn dealloc_res_id(&mut self, tid: usize) {
        self.task_res_allocator.lock().unwrap().dealloc(tid)
    }

    pub fn thread_count(&self) -> usize {
        self.tasks.len()
    }

    pub fn get_task(&self, tid: usize) -> Arc<TaskControlBlock> {
        self.tasks
            .iter()
            .find(|task| task.as_ref().unwrap().process.upgrade().unwrap().tid.0 == tid)
            .unwrap()
            .as_ref()
            .unwrap()
            .clone()
    }

    pub fn get_thread(&self, tid: usize) -> Arc<ProcessControlBlock> {
        self.thread_group
            .iter()
            .find(|proc| proc.tid.0 == tid)
            .unwrap()
            .clone()
    }
}

impl ProcessControlBlock {
    pub fn try_inner_exclusive_access(
        &self,
    ) -> Result<SpinMutexGuard<ProcessControlBlockInner>, ()> {
        self.inner.try_lock()
    }

    pub fn inner_lock_access(&self) -> Result<SpinMutexGuard<ProcessControlBlockInner>, ()> {
        self.inner.lock()
    }

    pub fn new(v: Arc<OSInode>) -> Arc<Self> {
        // memory_set with elf program headers/trampoline/trap context/user stack
        let (memory_set, ustack_base, entry_point, mut auxv) = MemorySet::from_elf_fd(v);
        print!("Entry point: {:#x}\r\n", entry_point);

        let new_token = memory_set.token();

        // allocate a pid
        let pid_handle = Arc::new(pid_alloc());
        let fdtable = FdTable::init();
        let process = Arc::new(Self {
            pid: pid_handle.clone(),
            tid: pid_handle.clone(),
            inner: SpinMutex::new(ProcessControlBlockInner {
                ctid: 0,
                is_zombie: false,
                exit_code: 0,
                parent: None,
                tasks: Vec::new(),
                children: Vec::new(),
                thread_group: Vec::new(),
                fd_table: Arc::new(SpinMutex::new(fdtable)),
                memory_set: Arc::new(SpinMutex::new(memory_set)),
                task_res_allocator: Arc::new(SpinMutex::new(RecycleAllocator::new(0))),
                mutex_list: Vec::new(),
                semaphore_list: Vec::new(),
                dir_entry: Some(root()),
                signal_manager: Arc::new(SpinMutex::new(SignalManager::new())),
            }),
        });
        // create a main thread, we should allocate ustack and trap_cx here
        let task = Arc::new(TaskControlBlock::new(
            Arc::clone(&process),
            ustack_base,
            true,
            CLONE_FLAGS::NONE,
        ));
        // prepare trap_cx of main thread
        let mut task_inner = task.inner_lock_access().unwrap();
        let trap_cx = task_inner.get_trap_cx();
        let ustack_top = task_inner.res.as_ref().unwrap().ustack_top();
        let kstack_top = task.kstack.get_top();

        let mut env = Vec::new();
        env.push(String::from("SHELL=/user_shell"));
        env.push(String::from("PWD=/"));
        env.push(String::from("USER=root"));
        env.push(String::from("MOTD_SHOWN=pam"));
        env.push(String::from("LANG=C.UTF-8"));
        env.push(String::from(
            "INVOCATION_ID=e9500a871cf044d9886a157f53826684",
        ));
        env.push(String::from("TERM=vt220"));
        env.push(String::from("SHLVL=2"));
        env.push(String::from("JOURNAL_STREAM=8:9265"));
        env.push(String::from("OLDPWD=/root"));
        env.push(String::from("_=busybox"));
        env.push(String::from("LOGNAME=root"));
        env.push(String::from("HOME=/"));
        env.push(String::from("PATH=/"));
        env.push(String::from("LD_LIBRARY_PATH=/"));
        

        let mut user_sp = task_inner.res.as_mut().unwrap().ustack_top();
        println!("User_sp {:#x}", user_sp);
        // envp[] is a null-terminated array of pointers to strings.
        let mut envp: Vec<usize> = (0..=env.len()).collect();
        envp[env.len()] = 0;
        for i in 0..env.len() {
            user_sp -= env[i].len() + 1;
            envp[i] = user_sp;
            let mut p = user_sp;
            for c in env[i].as_bytes() {
                *translated_refmut(new_token, p as *mut u8) = *c;
                p += 1;
            }
            *translated_refmut(new_token, p as *mut u8) = 0;
        }

        //argv[] 
        let args:Vec<String> = Vec::new();
        let mut argv: Vec<usize> = (0..=args.len()).collect();
        argv[args.len()] = 0;
        for i in 0..args.len() {
            user_sp -= args[i].len() + 1;
            argv[i] = user_sp;
            let mut p = user_sp;
            for c in args[i].as_bytes() {
                *translated_refmut(new_token, p as *mut u8) = *c;
                p += 1;
            }
            *translated_refmut(new_token, p as *mut u8) = 0;
        }

        // rand bytes
        user_sp -= 16;
        let mut tmp = user_sp;
        auxv.push(Auxv::new(AuxvType::AT_RANDOM, user_sp));
        for i in 0..16 {
            *translated_refmut(new_token, tmp as *mut u8) = i;
            tmp += 1;
        }

        // make the user_sp aligned to 8B for k210 platform
        user_sp -= user_sp % 16;

        // auxv[]
        auxv.push(Auxv::new(AuxvType::AT_EXECFN, argv[0]));
        auxv.push(Auxv::new(AuxvType::AT_NULL, 0));
        user_sp -= core::mem::size_of::<usize>() * 2 * auxv.len();
        auxv.iter().enumerate().for_each(|(idx, av)| {
            let addr = user_sp + core::mem::size_of::<usize>() * 2 * idx;
            *translated_refmut(new_token, (addr) as *mut usize) = av.a_type as usize;
            *translated_refmut(
                new_token,
                (addr + core::mem::size_of::<usize>()) as *mut usize,
            ) = av.val as usize;
        });

        // *envp[]
        user_sp -= (env.len() + 1) * core::mem::size_of::<usize>();
        envp.iter().enumerate().for_each(|(idx, ptr)| {
            *translated_refmut(
                new_token,
                (user_sp + idx * core::mem::size_of::<usize>()) as *mut usize,
            ) = *ptr;
        });

        // *argv[]
        user_sp -= (args.len() + 1) * core::mem::size_of::<usize>();
        argv.iter().enumerate().for_each(|(idx, ptr)| {
            *translated_refmut(
                new_token,
                (user_sp + idx * core::mem::size_of::<usize>()) as *mut usize,
            ) = *ptr;
        });

        // argc
        user_sp -= core::mem::size_of::<usize>();
        *translated_refmut(new_token, user_sp as *mut usize) = args.len();

        println!("usersp {:#x}",user_sp);
        drop(task_inner);
        *trap_cx = TrapContext::app_init_context(
            entry_point,
            user_sp,
            KERNEL_SPACE.lock().unwrap().token(),
            kstack_top,
            trap_handler as usize,
        );

        // add main thread to the process
        let mut process_inner = process.inner_lock_access().unwrap();
        process_inner.tasks.push(Some(Arc::clone(&task)));
        drop(process_inner);
        // add main thread to scheduler
        add_task(task);
        process
    }

    pub fn exec_fd(self: &Arc<Self>, args: Vec<String>, inode: Arc<OSInode>) {
        assert_eq!(self.inner_lock_access().unwrap().thread_count(), 1);

        // memory_set with elf program headers/trampoline/trap context/user stack
        let (memory_set, program_break, entry_point, mut auxv) = MemorySet::from_elf_fd(inode);

        let new_token = memory_set.token();

        // substitute memory_set
        self.inner_lock_access().unwrap().memory_set = Arc::new(SpinMutex::new(memory_set));
        // then we alloc user resource for main thread again
        // since memory_set has been changed
        let task = self.inner_lock_access().unwrap().get_task(self.tid.0);
        let mut task_inner = task.inner_lock_access().unwrap();

        task_inner.res.as_mut().unwrap().program_break = program_break;
        task_inner.res.as_mut().unwrap().alloc_user_res();
        task_inner.trap_cx_ppn = task_inner.res.as_mut().unwrap().trap_cx_ppn();

        let mut env = Vec::new();
        env.push(String::from("SHELL=/user_shell"));
        env.push(String::from("PWD=/"));
        env.push(String::from("USER=root"));
        env.push(String::from("MOTD_SHOWN=pam"));
        env.push(String::from("LANG=C.UTF-8"));
        env.push(String::from(
            "INVOCATION_ID=e9500a871cf044d9886a157f53826684",
        ));
        env.push(String::from("TERM=vt220"));
        env.push(String::from("SHLVL=2"));
        env.push(String::from("JOURNAL_STREAM=8:9265"));
        env.push(String::from("OLDPWD=/root"));
        env.push(String::from("_=busybox"));
        env.push(String::from("LOGNAME=root"));
        env.push(String::from("HOME=/"));
        env.push(String::from("PATH=/"));
        env.push(String::from("LD_LIBRARY_PATH=."));

        // push arguments on user stack
        let mut user_sp = task_inner.res.as_mut().unwrap().ustack_top();

        //envp[]
        let mut envp: Vec<usize> = (0..=env.len()).collect();
        envp[env.len()] = 0;
        for i in 0..env.len() {
            user_sp -= env[i].len() + 1;
            envp[i] = user_sp;
            let mut p = user_sp;
            for c in env[i].as_bytes() {
                *translated_refmut(new_token, p as *mut u8) = *c;
                p += 1;
            }
            *translated_refmut(new_token, p as *mut u8) = 0;
        }

        // argv[]
        let mut argv: Vec<usize> = (0..=args.len()).collect();
        argv[args.len()] = 0;
        for i in 0..args.len() {
            user_sp -= args[i].len() + 1;
            argv[i] = user_sp;
            let mut p = user_sp;
            for c in args[i].as_bytes() {
                *translated_refmut(new_token, p as *mut u8) = *c;
                p += 1;
            }
            *translated_refmut(new_token, p as *mut u8) = 0;
        }

        // rand bytes
        user_sp -= 16;
        let mut tmp = user_sp;
        auxv.push(Auxv::new(AuxvType::AT_RANDOM, user_sp));
        for i in 0..16 {
            *translated_refmut(new_token, tmp as *mut u8) = i;
            tmp += 1;
        }

        // make the user_sp aligned to 8B for k210 platform
        user_sp -= user_sp % 16;

        // auxv[]
        auxv.push(Auxv::new(AuxvType::AT_EXECFN, argv[0]));
        auxv.push(Auxv::new(AuxvType::AT_NULL, 0));
        user_sp -= core::mem::size_of::<usize>() * 2 * auxv.len();
        auxv.iter().enumerate().for_each(|(idx, av)| {
            let addr = user_sp + core::mem::size_of::<usize>() * 2 * idx;
            *translated_refmut(new_token, (addr) as *mut usize) = av.a_type as usize;
            *translated_refmut(
                new_token,
                (addr + core::mem::size_of::<usize>()) as *mut usize,
            ) = av.val as usize;
        });

        // *envp[]
        user_sp -= (env.len() + 1) * core::mem::size_of::<usize>();
        envp.iter().enumerate().for_each(|(idx, ptr)| {
            *translated_refmut(
                new_token,
                (user_sp + idx * core::mem::size_of::<usize>()) as *mut usize,
            ) = *ptr;
        });

        // *argv[]
        user_sp -= (args.len() + 1) * core::mem::size_of::<usize>();
        argv.iter().enumerate().for_each(|(idx, ptr)| {
            *translated_refmut(
                new_token,
                (user_sp + idx * core::mem::size_of::<usize>()) as *mut usize,
            ) = *ptr;
        });

        // argc
        user_sp -= core::mem::size_of::<usize>();
        *translated_refmut(new_token, user_sp as *mut usize) = args.len();

        // initialize trap_cx
        let mut trap_cx = TrapContext::app_init_context(
            entry_point,
            user_sp,
            KERNEL_SPACE.lock().unwrap().token(),
            task.kstack.get_top(),
            trap_handler as usize,
        );
        *task_inner.get_trap_cx() = trap_cx;
    }

    pub fn getpid(&self) -> usize {
        self.pid.0
    }

    pub fn gettid(&self) -> usize {
        self.tid.0
    }

    pub fn getppid(&self) -> usize {
        self.inner_lock_access()
            .unwrap()
            .parent
            .as_ref()
            .unwrap()
            .upgrade()
            .unwrap()
            .getpid()
    }
}

pub struct FdTable {
    pub fd_max: usize,
    pub tables: BTreeMap<usize, Option<Arc<dyn File + Send + Sync>>>,
}
impl FdTable {
    pub fn init() -> Self {
        let mut tables: BTreeMap<usize, Option<Arc<dyn File + Send + Sync>>> = BTreeMap::new();
        tables.insert(0, Some(Arc::new(Stdin)));
        tables.insert(1, Some(Arc::new(Stdout)));
        tables.insert(2, Some(Arc::new(Stdout)));
        Self {
            tables,
            fd_max: 1000,
        }
    }
    pub fn new() -> Self {
        Self {
            tables: BTreeMap::new(),
            fd_max: 1000,
        }
    }
    pub fn copy_from(old_fd: &Self) -> Self {
        let mut tables: BTreeMap<usize, Option<Arc<dyn File + Send + Sync>>> = BTreeMap::new();
        for (fd, opt) in old_fd.tables.iter() {
            if let Some(file) = opt {
                tables.insert(*fd, Some(file.clone()));
            } else {
                tables.insert(*fd, None);
            }
        }
        Self {
            tables,
            fd_max: old_fd.fd_max,
        }
    }
    pub fn alloc_fd(&mut self) -> usize {
        for i in 0..self.fd_max {
            if !self.tables.contains_key(&i) {
                self.tables.insert(i, None);
                return i;
            }
        }
        return usize::MAX;
    }
    pub fn get(&self, fd: usize) -> Option<&Option<Arc<dyn File + Send + Sync>>> {
        self.tables.get(&fd)
    }
}

pub fn do_fork(flags: usize, stack: usize, ptid: usize, tls: usize, ctid: usize) -> usize {
    let token = current_user_token();
    let process = current_process();
    let mut current = process.inner_lock_access().unwrap();

    let clone_flags = CLONE_FLAGS::from_bits(flags).unwrap();

    // copy mm
    let memory_set = match clone_flags.contains(CLONE_FLAGS::CLONE_VM) {
        true => current.memory_set.clone(),
        false => Arc::new(SpinMutex::new(MemorySet::from_user_cow(
            &mut current.memory_set.lock().unwrap(),
        ))),
    };


    //copy files
    let fd_table = match clone_flags.contains(CLONE_FLAGS::CLONE_FILES) {
        true => current.fd_table.clone(),
        false => Arc::new(SpinMutex::new(FdTable::copy_from(
            &current.fd_table.lock().unwrap(),
        ))),
    };

    // 复制或者生成新的PID以及TID
    let (pid, tid, task_res_allocator) = match clone_flags.contains(CLONE_FLAGS::CLONE_THREAD) {
        true => {
            let tid_handler = Arc::new(tid_alloc());
            (
                process.pid.clone(),
                tid_handler,
                current.task_res_allocator.clone(),
            )
        }
        false => {
            let pid_handler = Arc::new(pid_alloc());
            let task_res_allocator = Arc::new(SpinMutex::new(RecycleAllocator::new(0)));
            (pid_handler.clone(), pid_handler.clone(), task_res_allocator)
        }
    };

    let child = Arc::new(ProcessControlBlock {
        pid,
        tid,
        inner: SpinMutex::new(ProcessControlBlockInner {
            is_zombie: false,
            memory_set,
            parent: Some(Arc::downgrade(&process)),
            children: Vec::new(),
            thread_group: Vec::new(),
            exit_code: 0,
            fd_table,
            tasks: Vec::new(),
            task_res_allocator,
            mutex_list: Vec::new(),
            semaphore_list: Vec::new(),
            dir_entry: current.dir_entry.clone(),
            ctid,
            signal_manager: Arc::new(SpinMutex::new(SignalManager::new())),
        }),
    });

    if !clone_flags.contains(CLONE_FLAGS::CLONE_THREAD) {
        // 将新进程或者线程加入到当前进程的子进程列表中
        current.children.push(Arc::clone(&child));
    } else {
        // 将新进程或者线程加入到当前进程的线程组中
        current.thread_group.push(Arc::clone(&child));
    }

    let new_task = Arc::new(TaskControlBlock::new(
        Arc::clone(&child),
        current
            .get_task(process.tid.0)
            .inner_lock_access()
            .unwrap()
            .res
            .as_ref()
            .unwrap()
            .ustack_base(),
        // 当设置 CLONE_VM 时则分配USER_STACK以及TRAP——CONTEXT资源
        clone_flags.contains(CLONE_FLAGS::CLONE_VM),
        clone_flags,
    ));
    let mut child_inner = child.try_inner_exclusive_access().unwrap();
    child_inner.tasks.push(Some(Arc::clone(&new_task)));
    drop(child_inner);

    // 修改 Trap_Cx
    let new_task_inner = new_task.inner_lock_access().unwrap();
    let new_trap_cx = new_task_inner.get_trap_cx();
    new_trap_cx.x[10] = 0;
    new_trap_cx.kernel_sp = new_task.kstack.get_top();
    let tid = child.tid.0;

    if clone_flags.contains(CLONE_FLAGS::CLONE_PARENT_SETTID) {
        *translated_refmut(token, ptid as *mut usize) = tid;
    }
    if clone_flags.contains(CLONE_FLAGS::CLONE_CHILD_SETTID) {
        *translated_refmut(token, ctid as *mut usize) = tid;
    }
    if clone_flags.contains(CLONE_FLAGS::CLONE_VM) {
        let task = current_task().unwrap();
        let task_inner = task.inner_lock_access().unwrap();
        let new_task_res = new_task_inner.res.as_ref().unwrap();
        new_trap_cx.copy_from(task_inner.get_trap_cx(), new_task.kstack.get_top(), stack);
        (*new_trap_cx).x[10] = 0;
    }
    if clone_flags.contains(CLONE_FLAGS::CLONE_SETTLS) {
        (*new_trap_cx).x[4] = tls;
    }

    drop(new_task_inner);
    add_task(new_task);
    child.pid.0
}
