use super::{__switch, add_task};
use super::{fetch_task, TaskStatus};
use super::{ProcessControlBlock, TaskContext, TaskControlBlock};
use crate::config::MAX_HARTID;
use crate::sync::{intr_off, intr_on};
use crate::trap::TrapContext;
use alloc::collections::BTreeMap;
use alloc::sync::Arc;
use core::arch::asm;
pub static mut PROCESSORS: BTreeMap<usize, Processor> = BTreeMap::new();

pub struct Processor {
    pub noff: usize,
    pub intena: bool,
    current: Option<Arc<TaskControlBlock>>,
    idle_task_cx: TaskContext,
}

pub fn new() -> Processor {
    Processor {
        noff: 0,
        intena: false,
        current: None,
        idle_task_cx: TaskContext::zero_init(),
    }
}

impl Processor {
    pub fn new() -> Self {
        Self {
            noff: 0,
            intena: false,
            current: None,
            idle_task_cx: TaskContext::zero_init(),
        }
    }
    fn get_idle_task_cx_ptr(&mut self) -> *mut TaskContext {
        &mut self.idle_task_cx as *mut _
    }
    pub fn take_current(&mut self) -> Option<Arc<TaskControlBlock>> {
        self.current.take()
    }
    pub fn is_none(&self) -> bool {
        self.current.is_none()
    }
    pub fn is_some(&self) -> bool {
        self.current.is_some()
    }
    pub fn current(&self) -> Option<Arc<TaskControlBlock>> {
        self.current.as_ref().map(|task| Arc::clone(task))
    }
}

#[inline]
pub fn current_hartid() -> usize {
    let mut tp: usize;
    unsafe { asm!("mv {}, tp", out(reg) tp) };
    tp
}

pub fn init_hart() {
    unsafe {
        for hartid in 0..MAX_HARTID {
            PROCESSORS.insert(hartid, Processor::new());
        }
    }
}

pub fn run_tasks(hartid: usize) {
    let mut processor = current_processor().unwrap();
    loop {
        intr_on();
        intr_off();
        if let Some(task) = take_current_task() {
            add_task(task);
        }
        if let Some(task) = fetch_task() {
            intr_check!();
            assert!(current_hartid() == hartid);
            let mut task_inner = task.inner_lock_access().unwrap();
            let idle_task_cx_ptr = processor.get_idle_task_cx_ptr();
            let next_task_cx_ptr = &task_inner.task_cx as *const TaskContext;
            task_inner.task_status = TaskStatus::Running;
            drop(task_inner);
            processor.current = Some(task);
            assert!(processor.is_none() == false);
            unsafe {
                __switch(idle_task_cx_ptr, next_task_cx_ptr);
            }
        }
    }
}

#[inline]
pub fn current_processor() -> Option<&'static mut Processor> {
    unsafe { PROCESSORS.get_mut(&current_hartid()) }
}
#[inline]
pub fn take_current_task() -> Option<Arc<TaskControlBlock>> {
    let processor = current_processor().unwrap();
    processor.take_current()
}
#[inline]
pub fn current_task() -> Option<Arc<TaskControlBlock>> {
    let processor = current_processor().unwrap();
    processor.current()
}
#[inline]
pub fn current_process() -> Arc<ProcessControlBlock> {
    current_task().unwrap().process.upgrade().unwrap()
}
#[inline]
pub fn current_user_token() -> usize {
    let task = current_task().unwrap();
    let token = task.get_user_token();
    token
}
#[inline]
pub fn current_trap_cx() -> &'static mut TrapContext {
    current_task()
        .unwrap()
        .inner_lock_access()
        .unwrap()
        .get_trap_cx()
}
#[inline]
pub fn current_trap_cx_user_va() -> usize {
    current_task()
        .unwrap()
        .inner_lock_access()
        .unwrap()
        .res
        .as_ref()
        .unwrap()
        .trap_cx_user_va()
}
#[inline]
pub fn current_kstack_top() -> usize {
    current_task().unwrap().kstack.get_top()
}
#[inline]
pub fn enter_kernel() {
    current_task()
        .unwrap()
        .inner_lock_access()
        .unwrap()
        .enter_kernel();
}
#[inline]
pub fn exit_kernel() {
    current_task()
        .unwrap()
        .inner_lock_access()
        .unwrap()
        .exit_kernel();
}

pub fn schedule(switched_task_cx_ptr: *mut TaskContext) {
    intr_off();
    let processor = current_processor().unwrap();
    let idle_task_cx_ptr = processor.get_idle_task_cx_ptr();
    drop(processor);
    unsafe {
        __switch(switched_task_cx_ptr, idle_task_cx_ptr);
    }
}
