use super::context::Pthread;
use super::id::TaskUserRes;
use super::{kstack_alloc, KernelStack, ProcessControlBlock, TaskContext, CLONE_FLAGS};
use crate::mm::{
    translated_byte_buffer, translated_bytes, translated_copy_refmut, translated_ref,
    translated_refmut, UserBuffer,
};
use crate::sync::{SpinMutex, SpinMutexGuard};
use crate::timer::get_time_ms;
use crate::trap::TrapContext;
use crate::{mm::PhysPageNum, sync::UPSafeCell};
use alloc::slice;
use alloc::string::String;
use alloc::sync::{Arc, Weak};
use core::fmt::Debug;
use core::mem::size_of;
use core::ops::{Deref, DerefMut};
use core::ptr::NonNull;
use k210_pac::spi0::sr;
use log::info;
pub struct TaskControlBlock {
    // immutable
    pub process: Weak<ProcessControlBlock>,
    pub kstack: KernelStack,
    // mutable
    pub inner: SpinMutex<TaskControlBlockInner>,
}

impl TaskControlBlock {
    pub fn inner_lock_access(&self) -> Result<SpinMutexGuard<'_, TaskControlBlockInner>, ()> {
        self.inner.lock()
    }

    pub fn get_user_token(&self) -> usize {
        let process = self.process.upgrade().unwrap();
        let inner = process.inner_lock_access().unwrap();
        inner.get_user_token()
    }
}
pub struct TaskControlBlockInner {
    pub ticks: usize,
    pub times: Tms,
    pub res: Option<TaskUserRes>,
    pub trap_cx_ppn: PhysPageNum,
    pub task_cx: TaskContext,
    pub task_status: TaskStatus,
    pub exit_code: Option<i32>,
    pub flags: CLONE_FLAGS,
}

impl TaskControlBlockInner {
    pub fn get_trap_cx(&self) -> &'static mut TrapContext {
        self.trap_cx_ppn.get_mut()
    }

    pub fn backup_trap_cx(&self) {
        unsafe {
            let page = self.trap_cx_ppn.get_bytes_array().as_ptr() as *mut u8;
            let len = size_of::<TrapContext>();
            let dst = page.add(len);
            for i in 0..len {
                *dst.add(i) = *page.add(i);
            }
        }
    }

    pub fn restore_trap_cx(&self) {
        unsafe {
            let page = self.trap_cx_ppn.get_bytes_array().as_ptr() as *mut u8;
            let len = size_of::<TrapContext>();
            let src = page.add(len);
            for i in 0..len {
                *page.add(i) = *src.add(i);
            }
        }
    }

    #[allow(unused)]
    fn get_status(&self) -> TaskStatus {
        self.task_status
    }
    pub fn enter_kernel(&mut self) {
        let ticks = get_time_ms();
        self.times.tms_utime += ticks - self.ticks;
        self.ticks = ticks;
    }
    pub fn exit_kernel(&mut self) {
        let ticks = get_time_ms();
        self.times.tms_stime += ticks - self.ticks;
        self.ticks = ticks;
    }
}

impl TaskControlBlock {
    pub fn new(
        process: Arc<ProcessControlBlock>,
        program_beak: usize,
        alloc_user_res: bool,
        flags: CLONE_FLAGS,
    ) -> Self {
        let res = TaskUserRes::new(Arc::clone(&process), alloc_user_res, program_beak);
        let trap_cx_ppn = res.trap_cx_ppn();
        let kstack = kstack_alloc();
        let kstack_top = kstack.get_top();
        Self {
            process: Arc::downgrade(&process),
            kstack,
            inner: SpinMutex::new(TaskControlBlockInner {
                ticks: 0,
                times: Tms::default(),
                res: Some(res),
                trap_cx_ppn,
                task_cx: TaskContext::goto_trap_return(kstack_top),
                task_status: TaskStatus::Ready,
                exit_code: None,
                flags,
            }),
        }
    }
}

#[derive(Copy, Clone, PartialEq)]
pub enum TaskStatus {
    Ready,
    Waiting,
    Running,
    Blocking,
}

#[repr(C)]
#[derive(Default, Clone, Copy)]
pub struct Tms {
    tms_utime: usize,
    tms_stime: usize,
    tms_cutime: usize,
    tms_cstime: usize,
}

impl Deref for Tms {
    type Target = Self;

    fn deref(&self) -> &Self::Target {
        self
    }
}

impl DerefMut for Tms {
    fn deref_mut(&mut self) -> &mut Self::Target {
        self
    }
}

impl Debug for TaskControlBlock {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        f.write_str("taskcb")
    }
}

impl Debug for TaskControlBlockInner {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        f.write_str("taskcbinner")
    }
}
