mod context;
use crate::config::TRAMPOLINE;
use crate::irq::handler_ext;
use crate::mm::{do_page_fault, translated_ref};
use crate::sbi::sbi_smext_stimer;
use crate::sync::{handler_exist_signal, intr_off, intr_on, SpinMutex};
use crate::syscall::syscall;
use crate::task::{
    current_hartid, current_process, current_trap_cx, current_trap_cx_user_va, current_user_token,
    enter_kernel, exit_current_and_run_next, exit_kernel, suspend_current_and_run_next,
};
use crate::timer::{check_timer, set_next_trigger};
pub use context::TrapContext;
use log::info;
use riscv::asm;
use core::arch::{asm, global_asm};
use riscv::register::sstatus::{self, SPP};
use riscv::register::{
    mtvec::TrapMode,
    scause::{self, Exception, Interrupt, Trap},
    sie, stval, stvec,
};
use riscv::register::{sepc, sscratch};

global_asm!(include_str!("trap.S"));

pub fn init() {
    set_kernel_trap_entry();
}

fn set_kernel_trap_entry() {
    extern "C" {
        fn __alltraps();
        fn __alltraps_k();
    }
    let __alltraps_k_va = __alltraps_k as usize - __alltraps as usize + TRAMPOLINE;
    unsafe {
        stvec::write(__alltraps_k_va, TrapMode::Direct);
        sscratch::write(trap_from_kernel as usize);
    }
}

fn set_user_trap_entry() {
    unsafe {
        stvec::write(TRAMPOLINE as usize, TrapMode::Direct);
    }
}

pub fn enable_timer_interrupt() {
    unsafe {
        sie::set_stimer();
    }
}

lazy_static::lazy_static!(
    static ref KERNELLOCK:SpinMutex<()> = SpinMutex::new(());
);

#[no_mangle]
pub fn trap_handler() -> ! {
    debug_assert_eq!(sstatus::read().sie(), false, "Trap sie enabled");
    debug_assert_eq!(sstatus::read().spp(), SPP::User, "Trop not from User mode");
    set_kernel_trap_entry();
    let scause = scause::read();
    let stval = stval::read();
    enter_kernel();
    match scause.cause() {
        Trap::Exception(Exception::UserEnvCall) => {
            // jump to next instruction anyway
            let mut cx = current_trap_cx();

            // if cx.x[17] != 64 && cx.x[17] != 63 {
            //     println!("Syscall: {} {:#x}", cx.x[17], cx.sstatus.bits());
            //     println!("Spp {:?}", cx.sstatus.spp());
            // }

            cx.sepc += 4;
            // get system call return value
            let result = syscall(
                cx.x[17],
                [cx.x[10], cx.x[11], cx.x[12], cx.x[13], cx.x[14], cx.x[15]],
            );
            // cx is changed during sys_exec, so we have to call it again

            cx.x[10] = result as usize;
        }
        Trap::Exception(Exception::StoreFault)
        | Trap::Exception(Exception::StorePageFault)
        | Trap::Exception(Exception::InstructionFault)
        | Trap::Exception(Exception::InstructionPageFault)
        | Trap::Exception(Exception::LoadFault)
        | Trap::Exception(Exception::LoadPageFault) => {
            // Copy On Write 错页处理
            if let Err(_) = do_page_fault(current_trap_cx(), stval, scause.cause()) {
                exit_current_and_run_next(-2);
            }

            let trap_cx = current_trap_cx();
            trap_cx.sstatus.set_spp(SPP::User);
            // trap_cx.sepc += 4;
            
            // exit_current_and_run_next(-2);
        }
        Trap::Exception(Exception::IllegalInstruction) => {
            println!(
                "[kernel] IllegalInstruction in application, core dumped. stval {:#x}, sepc {:#x}",
                stval,
                current_trap_cx().sepc
            );
            // illegal instruction exit code
            exit_current_and_run_next(-3);
        }
        Trap::Interrupt(Interrupt::SupervisorTimer) => {
            set_next_trigger();
            if current_hartid() == 0 {
                check_timer();
            }
            handler_exist_signal();
            suspend_current_and_run_next();
        }
        Trap::Interrupt(Interrupt::SupervisorSoft) => {
            handler_ext();
            sbi_smext_stimer();
        }
        Trap::Interrupt(Interrupt::SupervisorExternal) => {
            
            handler_ext();
        }
        Trap::Exception(Exception::Breakpoint)=> {
            let token = current_user_token();
            let process = current_process();
            let trap_cx = current_trap_cx();
            trap_cx.sepc += 4;
        }
        _ => {
            panic!(
                "Unsupported U trap {:?}, stval = {:#x}!",
                scause.cause(),
                stval
            );
        }
    }

    trap_return();
}

#[no_mangle]
pub fn trap_return() -> ! {
    intr_off();
    set_user_trap_entry();
    exit_kernel();
    handler_exist_signal();
    let trap_cx_user_va = current_trap_cx_user_va();
    let user_satp = current_user_token();
    extern "C" {
        fn __alltraps();
        fn __restore();
    }
    let restore_va = __restore as usize - __alltraps as usize + TRAMPOLINE;

    unsafe {
        asm!(
            "fence.i",
            "jr {restore_va}",
            restore_va = in(reg) restore_va,
            in("a0") trap_cx_user_va,
            in("a1") user_satp,
            options(noreturn)
        );
    }
}

#[no_mangle]
pub fn trap_from_kernel() {
    assert_eq!(sstatus::read().spp(), SPP::Supervisor);
    debug_assert_eq!(sstatus::read().sie(), false, "Trap_from_kernel sie enabled");
    let scause = scause::read();
    let stval = stval::read();
    let sepc = sepc::read();
    match scause.cause() {
        Trap::Interrupt(Interrupt::SupervisorSoft) => {
            handler_ext();
            sbi_smext_stimer();
        }
        Trap::Interrupt(Interrupt::SupervisorTimer) => {
            set_next_trigger();
            if current_hartid() == 0 {
                check_timer();
            }
        }
        Trap::Interrupt(Interrupt::SupervisorExternal) => {
            handler_ext();
        }
        _ => {
            panic!(
                "Unsupported S trap {:?}, stval = {:#x}, sepc = {:#x}",
                scause.cause(),
                stval,
                sepc,
            );
        }
    }
}
